# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import os
import pandas as pd
from tqdm import tqdm
from datetime import datetime

from ..pipelines.classify_rascore import classify_rascore

from ..constants.conf import sw1_name, sw2_name, y32_name, y71_name, loop_resid_dict
from ..constants.nuc import nuc_class_dict, gtp_name
from ..constants.gene import uniprot_acc_lst, swiss_id_lst, gene_class_dict
from ..constants.mut import mut_class_lst, other_mut_name
from ..constants.pharm import (
    pharm_match_dict,
    pharm_site_dict,
    sp2_name,
    sp12_name,
    mult_pharm_name,
    none_pharm_name,
    other_pharm_name,
)
from ..constants.dimer import dimer_name, none_dimer_name
from ..constants.prot import (
    gef_name,
    prot_pfam_dict,
    prot_class_dict,
    rem_name,
    cdc_name,
    nano_name,
    mult_prot_name,
)
from ..constants.pml import sup_pdb_code, sup_chainid


from ..scripts.search_pdbaa import search_pdbaa
from ..scripts.prep_coord import prep_coord
from ..scripts.prep_dih import prep_dih
from ..scripts.annot_mut import annot_mut
from ..scripts.annot_lig import annot_lig
from ..scripts.annot_prot import annot_prot
from ..scripts.annot_cf import annot_cf
from ..scripts.prep_interf import prep_interf
from ..scripts.build_interf_table import build_interf_table
from ..scripts.prep_pocket import prep_pocket, pocket_bound_name, pocket_unbound_name
from ..scripts.build_pocket_table import build_pocket_table

from ..functions.table import (
    mask_equal,
    make_dict,
    merge_dicts,
    mask_unequal,
    lst_col,
    title_str,
)
from ..functions.path import (
    get_renum_path,
    get_dir_path,
    copy_path,
    load_json,
    save_table,
    save_json,
    get_file_path,
    load_table,
    get_neighbor_path,
    delete_path,
    get_file_name,
    pipelines_str,
    data_str,
    classify_str,
    rascore_str,
    build_str,
    pdbaa_str,
)
from ..functions.lst import (
    lst_to_str,
    str_to_lst,
)
from ..functions.file import (
    entry_table_file,
    sifts_json_file,
    dih_json_file,
    interf_json_file,
    interf_table_file,
    pocket_json_file,
    pocket_table_file,
    result_table_file,
    pdbaa_fasta_file,
)
from ..functions.url import pdbaa_url
from ..functions.download import download_unzip
from ..functions.col import (
    pdb_id_col,
    chainid_col,
    core_path_col,
    bound_lig_col,
    bound_prot_chainid_col,
    mut_class_col,
    pharm_class_col,
    prot_class_col,
    cf_col,
    mut_status_col,
    pharm_lig_site_col,
    pharm_lig_col,
    pharm_lig_match_col,
    match_class_col,
    gene_class_col,
    prot_class_col,
    prot_col,
    bio_lig_col,
    nuc_class_col,
    rcsb_path_col,
    bound_prot_pfam_col,
    complete_col,
    interf_class_col,
    method_col,
    method_col,
    renum_path_col,
    pocket_class_col,
    pocket_type_col,
    pocket_status_col,
    pocket_site_col,
    pocket_id_col,
    pocket_cont_col,
)


def update_prep(out_path=None, past_df=None, num_cpu=1):

    entry_table_path = get_file_path(entry_table_file, dir_path=out_path)
    sifts_json_path = get_file_path(sifts_json_file, dir_path=out_path)
    dih_json_path = get_file_path(dih_json_file, dir_path=out_path)

    df = load_table(entry_table_path)

    delete_paths = True
    if past_df is not None:
        if (
            len(
                [
                    x
                    for x in [bound_lig_col, bound_prot_chainid_col]
                    if x not in list(past_df.columns)
                ]
            )
            == 0
        ):
            df = mask_unequal(df, pdb_id_col, lst_col(past_df, pdb_id_col))
            delete_paths = False

    if delete_paths:
        delete_path(sifts_json_path)
        delete_path(dih_json_path)

    if len(df) > 0:
        pdb_id_lst = lst_col(df, pdb_id_col, unique=True)

        past_sifts_dict = load_json(sifts_json_path)

        prep_coord(
            pdb_id_lst=pdb_id_lst,
            renum_script_path=get_neighbor_path(__file__, pipelines_str, "PDBrenum")
            + "/PDBrenum.py",
            coord_table_path=entry_table_path,
            core_dir=out_path,
            rcsb_dir=out_path,
            sifts_dir=out_path,
            renum_dir=out_path,
            sifts_json_path=sifts_json_path,
            data=df,
            update_coords=False,
            num_cpu=num_cpu,
        )

        sifts_dict = load_json(sifts_json_path)
        if past_sifts_dict is not None:
            sifts_dict = merge_dicts([sifts_dict, past_sifts_dict])
        save_json(sifts_json_path, sifts_dict)
        
        try:
            df = load_table(entry_table_path)

            core_path_lst = lst_col(df, core_path_col, unique=True)
            dih_dict = prep_dih(coord_paths=core_path_lst, num_cpu=num_cpu)

            past_dih_dict = load_json(dih_json_path)
            if past_dih_dict is not None:
                dih_dict = merge_dicts([dih_dict, past_dih_dict])
            save_json(dih_json_path, dih_dict)

            if past_df is not None:
                if (
                    len(
                        [
                            x
                            for x in [bound_lig_col, bound_prot_chainid_col]
                            if x not in list(past_df.columns)
                        ]
                    )
                    == 0
                ):
                    df = pd.concat([df, past_df], sort=False)
                    save_table(entry_table_path, df)
        except:
            if past_df is not None:
                save_table(entry_table_path, past_df)



def update_annot(pdbaa_fasta_path, out_path=None, past_df=None, num_cpu=1):

    entry_table_path = get_file_path(entry_table_file, dir_path=out_path)

    df = load_table(entry_table_path)

    if past_df is not None:
        if (
            len(
                [
                    x
                    for x in [
                        gene_class_col,
                        nuc_class_col,
                        mut_class_col,
                        pharm_class_col,
                        match_class_col,
                        prot_class_col,
                        cf_col,
                    ]
                    if x not in list(past_df.columns)
                ]
            )
            == 0
        ):
            df = mask_unequal(df, pdb_id_col, lst_col(past_df, pdb_id_col))

    if len(df) > 0:
        df = annot_mut(
            df=df,
            uniprot_accs=uniprot_acc_lst,
            resids="1-166",
            seq_dir=out_path,
            num_cpu=num_cpu,
        )

        for index in list(df.index.values):
            df.at[index, mut_class_col] = lst_to_str(
                [
                    x
                    for x in mut_class_lst
                    if x in str_to_lst(df.at[index, mut_status_col])
                ],
                empty=other_mut_name,
            )

        df = annot_lig(
            df=df,
            lig_dir=out_path,
            site_dict=pharm_site_dict,
            match_dict=pharm_match_dict,
            num_cpu=num_cpu,
        )

        for index in list(df.index.values):
            pharm_class = other_pharm_name
            pharm_site = df.at[index, pharm_lig_site_col]
            if "|" in pharm_site:
                pharm_class = mult_pharm_name
            elif pharm_site in [sp2_name, sp12_name, none_pharm_name]:
                pharm_class = pharm_site
            df.at[index, pharm_class_col] = pharm_class

        for index in list(df.index.values):
            pharm_match = df.at[index, pharm_lig_match_col]
            if "|" in pharm_match:
                match_class = mult_pharm_name
            else:
                match_class = pharm_match
            df.at[index, match_class_col] = match_class

        df[gene_class_col] = df[prot_col].map(gene_class_dict)
        df[nuc_class_col] = df[bio_lig_col].map(nuc_class_dict).fillna(gtp_name)

        df = annot_prot(
            df=df,
            pdbaa_fasta_path=pdbaa_fasta_path,
            pfam_dict=prot_pfam_dict,
        )

        for index in list(df.index.values):
            prot_pfam = df.at[index, bound_prot_pfam_col]
            if "|" in prot_pfam:
                if nano_name in prot_pfam:
                    prot_class = nano_name
                else:
                    prot_class = mult_prot_name
            else:
                for prot_name, pfam_lst in prot_class_dict.items():
                    if prot_pfam in pfam_lst:
                        prot_class = prot_name
                if prot_class == gef_name:
                    nuc_class = df.at[index, nuc_class_col]
                    prot_class += "."
                    if nuc_class == gtp_name:
                        prot_class += rem_name
                    else:
                        prot_class += cdc_name
            df.at[index, prot_class_col] = prot_class

        if (
            past_df is not None
            and len([x for x in [
                        gene_class_col,
                        nuc_class_col,
                        mut_class_col,
                        pharm_class_col,
                        match_class_col,
                        prot_class_col,
                        cf_col,
                    ] if x not in list(past_df.columns)]) == 0
        ):
            df[complete_col] = str(False)
            df = pd.concat([df, past_df], sort=False)
            save_table(entry_table_path, df)

        rcsb_path_lst = lst_col(df, rcsb_path_col, unique=True)
        df = annot_cf(
            coord_paths=rcsb_path_lst,
            data=df,
            num_cpu=num_cpu,
        )

        save_table(entry_table_path, df)


def update_interf(out_path=None, past_df=None, num_cpu=1):

    entry_table_path = get_file_path(entry_table_file, dir_path=out_path)
    interf_json_path = get_file_path(interf_json_file, dir_path=out_path)
    interf_table_path = get_file_path(interf_table_file, dir_path=out_path)

    df = load_table(entry_table_path)

    delete_paths = True
    if past_df is not None and interf_class_col in list(past_df.columns):
        df = mask_unequal(df, pdb_id_col, lst_col(past_df, pdb_id_col))
        delete_paths = False

    if delete_paths:
        delete_path(interf_json_path)
        delete_path(interf_table_path)

    if len(df) > 0:
        xray_df = mask_equal(df, method_col, "XRAY")

        coord_path_lst = lst_col(xray_df, renum_path_col, unique=True)

        chainid_dict = dict()
        for coord_path in coord_path_lst:
            chainid_dict[coord_path] = list()
        for index in list(xray_df.index.values):
            coord_path = xray_df.at[index, renum_path_col]
            chainid = xray_df.at[index, chainid_col]
            chainid_dict[coord_path].append(chainid)

        interf_dict = prep_interf(
            coord_paths=coord_path_lst,
            interf_dir=out_path,
            chainid_dict=chainid_dict,
            num_cpu=num_cpu,
        )

        past_interf_dict = load_json(interf_json_path)
        if past_interf_dict is not None:
            interf_dict = merge_dicts([interf_dict, past_interf_dict])
        save_json(interf_json_path, interf_dict)

        interf_df = build_interf_table(
            df=xray_df,
            interf_dict=interf_dict,
            search_coord_path=get_renum_path(sup_pdb_code, dir_path=out_path),
            search_chainid=sup_chainid,
            search_interf="4",
        )

        past_interf_df = load_table(interf_table_path)
        if past_interf_df is not None and interf_class_col in list(past_df.columns):
            interf_df = pd.concat([interf_df, past_interf_df], sort=False)
        save_table(interf_table_path, interf_df)

        interf_pdb_id_lst = lst_col(interf_df, pdb_id_col)
        df[interf_class_col] = (
            df[pdb_id_col]
            .map(make_dict(interf_pdb_id_lst, len(interf_pdb_id_lst) * [dimer_name]))
            .fillna(none_dimer_name)
        )

        if past_df is not None:
            if interf_class_col in list(past_df.columns):
                df[complete_col] = str(False)
                df = pd.concat([df, past_df], sort=False)

        save_table(entry_table_path, df)


def update_pocket(out_path=None, past_df=None, num_cpu=1):

    entry_table_path = get_file_path(entry_table_file, dir_path=out_path)
    pocket_json_path = get_file_path(pocket_json_file, dir_path=out_path)
    pocket_table_path = get_file_path(pocket_table_file, dir_path=out_path)

    df = load_table(entry_table_path)

    delete_paths = True
    if past_df is not None:
        if pocket_class_col in list(past_df.columns):
            df = mask_unequal(df, pdb_id_col, lst_col(past_df, pdb_id_col))
            delete_paths = False

    if delete_paths:
        delete_path(pocket_json_path)
        delete_path(pocket_table_path)

    if len(df) > 0:
        coord_path_lst = [x.replace(".cif", ".pdb") for x in lst_col(df, core_path_col)]

        chainid_dict = dict()
        for coord_path in coord_path_lst:
            chainid_dict[coord_path] = list()
        for index in list(df.index.values):
            coord_path = df.at[index, core_path_col].replace(".cif", ".pdb")
            chainid = df.at[index, chainid_col]
            chainid_dict[coord_path].append(chainid)

        pocket_dict = prep_pocket(
            coord_paths=coord_path_lst,
            pocket_dir=out_path,
            chainid_dict=chainid_dict,
            num_cpu=num_cpu,
        )

        past_pocket_dict = load_json(pocket_json_path)
        if past_pocket_dict is not None:
            pocket_dict = merge_dicts([pocket_dict, past_pocket_dict])
        save_json(pocket_json_path, pocket_dict)

        if past_df is not None:
            if pocket_class_col in list(past_df.columns):
                df[complete_col] = str(False)
                df = pd.concat([df, past_df], sort=False)

        temp_df = build_pocket_table(
            df=df,
            pocket_dict=pocket_dict,
        )
        temp_df = mask_equal(
            temp_df, pocket_type_col, [title_str(pharm_lig_col), pocket_unbound_name]
        )

        pocket_df = pd.DataFrame()
        for pharm_name in [sp2_name, sp12_name]:
            bound_df = mask_equal(temp_df, pharm_class_col, pharm_name)
            bound_df = mask_equal(bound_df, pocket_status_col, pocket_bound_name)

            bound_df[pocket_type_col] = pharm_name
            bound_df[pocket_site_col] = pharm_name
            pocket_df = pd.concat([pocket_df, bound_df], sort=False)

            search_cont_lst = lst_col(bound_df, pocket_cont_col)
            search_cont_lst = [str_to_lst(x) for x in search_cont_lst]

            unbound_df = build_pocket_table(
                df=mask_equal(df, pharm_class_col, none_pharm_name),
                pocket_dict=pocket_dict,
                search_cont_lst=search_cont_lst,
                search_max_dist=0.4,
                use_simpson=True,
            )

            if len(unbound_df) > 0:
                unbound_df = mask_equal(
                    unbound_df, pocket_status_col, pocket_unbound_name
                )
                unbound_df[pocket_site_col] = pharm_name
                pocket_df = pd.concat([pocket_df, unbound_df], sort=False)

        other_df = mask_unequal(
            temp_df, pocket_id_col, lst_col(pocket_df, pocket_id_col)
        )
        other_df = mask_equal(other_df, pocket_status_col, pocket_unbound_name)
        other_df = mask_equal(other_df, pharm_class_col, none_pharm_name)

        other_df[pocket_site_col] = other_pharm_name

        pocket_df = pd.concat([pocket_df, other_df], sort=False)

        save_table(pocket_table_path, pocket_df)

        pocket_df = load_table(pocket_table_path)

        annot_df = mask_equal(pocket_df, pocket_site_col, [sp2_name, sp12_name])

        for index in tqdm(
            list(df.index.values),
            desc="Annotating pockets",
            position=0,
            leave=True,
        ):
            pdb_df = mask_equal(annot_df, pdb_id_col, df.at[index, pdb_id_col])

            pocket_site_lst = lst_col(pdb_df, pocket_site_col, unique=True)
            pocket_status_lst = lst_col(pdb_df, pocket_status_col, unique=True)

            if len(pocket_site_lst) > 1:
                pocket_site_lst = [mult_pharm_name]

            if len(pocket_site_lst) == 0:
                pharm_class = df.at[index, pharm_class_col]
                if pharm_class == none_pharm_name:
                    pocket_site_lst = [other_pharm_name]
                    pocket_status_lst = [pocket_unbound_name]
                else:
                    pocket_site_lst = [pharm_class]
                    pocket_status_lst = [pocket_bound_name]

            pocket_site = lst_to_str(pocket_site_lst, join_txt="|")
            pocket_status = lst_to_str(pocket_status_lst)

            df.at[index, pocket_site_col] = pocket_site
            df.at[index, pocket_status_col] = pocket_status

            pocket_class = pocket_site

            pocket_class += "-"
            pocket_class += pocket_status

            df.at[index, pocket_class_col] = pocket_class

        save_table(entry_table_path, df)


def update_classify(out_path=None, past_df=None, num_cpu=1):

    entry_table_path = get_file_path(entry_table_file, dir_path=out_path)
    dih_json_path = get_file_path(dih_json_file, dir_path=out_path)

    df = load_table(entry_table_path)

    if past_df is not None:
        if (
            len(
                [
                    x
                    for x in [
                        sw1_name,
                        sw2_name,
                        y32_name,
                        y71_name
                    ]
                    if x not in list(past_df.columns)
                ]
            )
            == 0
        ):
            df = mask_unequal(df, pdb_id_col, lst_col(past_df, pdb_id_col))

    dih_dict = load_json(dih_json_path)

    classify_path = f"{out_path}/{classify_str}"

    if len(df) > 0:
        classify_rascore(
            df,
            out_path=classify_path,
            dih_dict=dih_dict,
            num_cpu=num_cpu,
        )

        result_df = load_table(get_file_path(result_table_file, dir_path=classify_path))

        for col in [sw1_name, sw2_name, y32_name, y71_name]:
            df[col] = df[pdb_id_col].map(
                make_dict(lst_col(result_df, pdb_id_col), lst_col(result_df, col))
            )

        if past_df is not None:
            if (
                len(
                    [
                        x
                        for x in [
                            sw1_name,
                            sw2_name,
                            y32_name,
                            y71_name
                        ]
                        if x not in list(past_df.columns)
                    ]
                )
                == 0
            ):
                df[complete_col] = str(False)
                df = pd.concat([df, past_df], sort=False)

        save_table(entry_table_path, df)


def build_rascore(out_path=None, pdbaa_fasta_path=None, num_cpu=1):

    if out_path is None:
        out_path = f"{os.getcwd()}/{rascore_str}_{build_str}"

    entry_table_path = get_file_path(entry_table_file, dir_path=out_path)
    pocket_table_path = get_file_path(pocket_table_file, dir_path=out_path)
    interf_table_path = get_file_path(interf_table_file, dir_path=out_path)
    dih_json_path = get_file_path(dih_json_file, dir_path=out_path)

    if pdbaa_fasta_path is None:
        curr_date = datetime.today().strftime("%Y-%m-%d")
        pdbaa_fasta_path = get_file_path(
            f"{curr_date}_{pdbaa_fasta_file}",
            dir_str=pdbaa_str,
            dir_path=out_path,
            pre_str=False,
        )

    past_df = load_table(entry_table_path)

    if past_df is None:
        print("No rascore database found! Building rascore database from scratch!")
    else:
        if complete_col in list(past_df.columns):
            past_df = mask_equal(past_df, complete_col, str(True))
        else:
            past_df[complete_col] = str(True)

    print("Downloading updated pdbaa file.")

    try:
        download_unzip(url=pdbaa_url, path=pdbaa_fasta_path)
    except:
        pdbaa_fasta_path_lst = [
            x
            for x in os.listdir(get_dir_path(dir_str=pdbaa_str, dir_path=out_path))
            if pdbaa_fasta_file in x
        ]
        if len(pdbaa_fasta_path_lst) > 0:
            pdbaa_fasta_path = sorted(
                pdbaa_fasta_path_lst,
                reverse=True,
            )[0]

            last_date = pdbaa_fasta_path.split(f"_{pdbaa_fasta_file}")[0]
            last_date = last_date.split(f"{pdbaa_str}/")[1]
            print(f"Updated pdbaa file unavailable. Using latest version ({last_date})")

    try:
        search_pdbaa(
            pdbaa_fasta_path=pdbaa_fasta_path,
            search_lst=swiss_id_lst,
            entry_table_path=entry_table_path,
            min_length=25,
        )
    except:
        print("Error reading pdbaa file.")
        copy_path(
            get_file_path(
                entry_table_file,
                dir_path=get_neighbor_path(__file__, pipelines_str, data_str),
            ),
            entry_table_path,
        )

    if past_df is not None:
        df = load_table(entry_table_path)
        if len(df) < len(past_df):
            print("Downgrading rascore database to older version.")
            past_df = None

    update_prep(out_path=out_path, past_df=past_df, num_cpu=num_cpu)
    update_annot(
        pdbaa_fasta_path,
        out_path=out_path,
        past_df=past_df,
        num_cpu=num_cpu,
    )
    update_interf(out_path=out_path, past_df=past_df, num_cpu=num_cpu)
    update_pocket(out_path=out_path, past_df=past_df, num_cpu=num_cpu)
    update_classify(out_path=out_path, past_df=past_df, num_cpu=num_cpu)

    df = load_table(entry_table_path)
    if complete_col in list(df.columns):
        del df[complete_col]

    save_table(entry_table_path, df)

    ref_df = df.set_index(pdb_id_col)

    pocket_df = load_table(pocket_table_path)
    interf_df = load_table(pocket_table_path)

    for col in list(ref_df.columns):
        if col not in list(pocket_df.columns):
            for index in list(pocket_df.index.values):
                pocket_df.at[index, col] = ref_df.at[pocket_df.at[index, pdb_id_col], col]
        if col not in list(interf_df.columns):
            for index in list(interf_df.index.values):
                interf_df.at[index, col] = ref_df.at[interf_df.at[index, pdb_id_col], col]

    save_table(pocket_table_path, pocket_df)
    save_table(interf_table_path, interf_df)

    for file_path in [
        entry_table_path,
        pocket_table_path,
        interf_table_path,
        dih_json_path,
    ]:
        copy_path(
            file_path,
            get_file_path(
                get_file_name(file_path),
                dir_path=get_neighbor_path(__file__, pipelines_str, data_str),
            ),
        )

    print("Rascore update complete!")


from src.download.modules import *
from src.download import lefttodownload, catalogdownloader, lookfilesinside, latestcatreader
from src.download.downloadwithThreadPool import run_downloads_with_ThreadPool, url_formation_for_pool, download_pdb_assemblies_list_with_lxml


# default_input_path_to_mmCIF = current_directory + "/mmCIF"
# default_input_path_to_PDB = current_directory + "/PDB"
# default_input_path_to_SIFTS = current_directory + "/SIFTS"
# default_output_path_to_mmCIF = current_directory + "/output_mmCIF"
# default_output_path_to_PDB = current_directory + "/output_PDB"
#
# default_input_path_to_mmCIF_assemblies = current_directory + "/mmCIF_assembly"
# default_input_path_to_PDB_assemblies = current_directory + "/PDB_assembly"
# default_output_path_to_mmCIF_assemblies = current_directory + "/output_mmCIF_assembly"
# default_output_path_to_PDB_assemblies = current_directory + "/output_PDB_assembly"


def supreme_download_master(format_of_db, job_type=None,
                            default_input_path_to_mmCIF=current_directory + "/mmCIF",
                            default_input_path_to_PDB=current_directory + "/PDB",
                            default_input_path_to_SIFTS=current_directory + "/SIFTS",
                            default_input_path_to_mmCIF_assembly=current_directory + "/mmCIF_assembly",
                            default_input_path_to_PDB_assembly=current_directory + "/PDB_assembly",

                            default_output_path_to_mmCIF=current_directory + "/output_mmCIF",
                            default_output_path_to_PDB=current_directory + "/output_PDB",
                            default_output_path_to_mmCIF_assemblies=current_directory + "/output_mmCIF_assembly",
                            default_output_path_to_PDB_assemblies=current_directory + "/output_PDB_assembly"):

    catalogdownloader.catalog_downloader()

    if job_type == "refresh":
        if os.path.exists(default_input_path_to_SIFTS):
            shutil.rmtree(default_input_path_to_SIFTS)
        if format_of_db == "mmCIF":
            if os.path.exists(default_input_path_to_mmCIF):
                shutil.rmtree(default_input_path_to_mmCIF)
            if os.path.exists(default_output_path_to_mmCIF):
                shutil.rmtree(default_output_path_to_mmCIF)

        if format_of_db == "mmCIF_assembly":
            if os.path.exists(default_input_path_to_mmCIF_assembly):
                shutil.rmtree(default_input_path_to_mmCIF_assembly)
            if os.path.exists(default_output_path_to_mmCIF_assemblies):
                shutil.rmtree(default_output_path_to_mmCIF_assemblies)

        if format_of_db == "PDB":
            if os.path.exists(default_input_path_to_PDB):
                shutil.rmtree(default_input_path_to_PDB)
            if os.path.exists(default_output_path_to_PDB):
                shutil.rmtree(default_output_path_to_PDB)

        if format_of_db == "PDB_assembly":
            if os.path.exists(default_input_path_to_PDB_assembly):
                shutil.rmtree(default_input_path_to_PDB_assembly)
            if os.path.exists(default_output_path_to_PDB_assemblies):
                shutil.rmtree(default_output_path_to_PDB_assemblies)

        if format_of_db == "all":
            if os.path.exists(default_input_path_to_PDB):
                shutil.rmtree(default_input_path_to_PDB)
            if os.path.exists(default_input_path_to_mmCIF):
                shutil.rmtree(default_input_path_to_mmCIF)
            if os.path.exists(default_input_path_to_PDB_assembly):
                shutil.rmtree(default_input_path_to_PDB_assembly)
            if os.path.exists(default_input_path_to_mmCIF_assembly):
                shutil.rmtree(default_input_path_to_mmCIF_assembly)

            if os.path.exists(default_output_path_to_mmCIF):
                shutil.rmtree(default_output_path_to_mmCIF)
            if os.path.exists(default_output_path_to_mmCIF_assemblies):
                shutil.rmtree(default_output_path_to_mmCIF_assemblies)
            if os.path.exists(default_output_path_to_PDB):
                shutil.rmtree(default_output_path_to_PDB)
            if os.path.exists(default_output_path_to_PDB_assemblies):
                shutil.rmtree(default_output_path_to_PDB_assemblies)

    if format_of_db == "mmCIF":
        all_data_from_catreader = latestcatreader.latest_catalog_reader()
        all_mmCIF_files_from_latest_catalog = all_data_from_catreader[0]
        all_SIFTS_files_from_latest_catalog = all_data_from_catreader[2]

        input_mmCIF_files_were_found = lookfilesinside.look_what_is_inside("mmCIF", default_input_path_to_mmCIF=default_input_path_to_mmCIF)
        left_to_download_mmCIF = lefttodownload.what_is_left_to_download(input_mmCIF_files_were_found, all_mmCIF_files_from_latest_catalog)
        urls_to_target_mmCIF_files = url_formation_for_pool("mmCIF", left_to_download_mmCIF, default_input_path_to_mmCIF=default_input_path_to_mmCIF)
        run_downloads_with_ThreadPool("mmCIF", urls_to_target_mmCIF_files, default_input_path_to_mmCIF=default_input_path_to_mmCIF)

        input_SIFTS_files_were_found = lookfilesinside.look_what_is_inside("SIFTS", default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        left_to_download_SIFTS = lefttodownload.what_is_left_to_download(input_SIFTS_files_were_found, all_SIFTS_files_from_latest_catalog)
        urls_to_target_SIFTS_files = url_formation_for_pool("SIFTS", left_to_download_SIFTS, default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        run_downloads_with_ThreadPool("SIFTS", urls_to_target_SIFTS_files, default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        return left_to_download_mmCIF

    if format_of_db == "mmCIF_assembly":
        all_data_from_catreader = latestcatreader.latest_catalog_reader()
        all_mmCIF_files = all_data_from_catreader[0]
        all_SIFTS_files_from_latest_catalog = all_data_from_catreader[2]

        lefttodownload_mmCIF_assemblies = list()
        input_mmCIF_assembly_files_were_found = lookfilesinside.look_what_is_inside(
            "mmCIF_assembly", default_input_path_to_mmCIF_assembly=default_input_path_to_mmCIF_assembly)

        all_mmCIF_files_4char = set()
        for mmCIF_file in all_mmCIF_files:
            all_mmCIF_files_4char.add(mmCIF_file[:4])

        input_mmCIF_assembly_files_were_found_4char = set()
        for mmCIF_assembly_file in input_mmCIF_assembly_files_were_found:
            input_mmCIF_assembly_files_were_found_4char.add(mmCIF_assembly_file[:4])

        set_difference = all_mmCIF_files_4char - input_mmCIF_assembly_files_were_found_4char
        list_difference = list(set_difference)

        for mmCIF_id in list_difference:
            lefttodownload_mmCIF_assemblies.append(mmCIF_id + ".cif.gz")

        urls_to_target_mmCIF_assembly_files = url_formation_for_pool("mmCIF_assembly", lefttodownload_mmCIF_assemblies)
        run_downloads_with_ThreadPool("mmCIF_assembly", urls_to_target_mmCIF_assembly_files,
                                      default_input_path_to_mmCIF_assembly=default_input_path_to_mmCIF_assembly)

        input_SIFTS_files_were_found = lookfilesinside.look_what_is_inside("SIFTS", default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        left_to_download_SIFTS = lefttodownload.what_is_left_to_download(input_SIFTS_files_were_found, all_SIFTS_files_from_latest_catalog)
        urls_to_target_SIFTS_files = url_formation_for_pool("SIFTS", left_to_download_SIFTS, default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        run_downloads_with_ThreadPool("SIFTS", urls_to_target_SIFTS_files, default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        return lefttodownload_mmCIF_assemblies

    if format_of_db == "PDB":
        all_data_from_catreader = latestcatreader.latest_catalog_reader()
        all_PDB_files_from_latest_catalog = all_data_from_catreader[1]
        all_SIFTS_files_from_latest_catalog = all_data_from_catreader[2]

        input_PDB_files_were_found = lookfilesinside.look_what_is_inside("PDB", default_input_path_to_PDB=default_input_path_to_PDB)
        left_to_download_PDB = lefttodownload.what_is_left_to_download(input_PDB_files_were_found, all_PDB_files_from_latest_catalog)
        urls_to_target_PDB_files = url_formation_for_pool("PDB", left_to_download_PDB, default_input_path_to_PDB=default_input_path_to_PDB)
        run_downloads_with_ThreadPool("PDB", urls_to_target_PDB_files, default_input_path_to_PDB=default_input_path_to_PDB)

        input_SIFTS_files_were_found = lookfilesinside.look_what_is_inside("SIFTS", default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        left_to_download_SIFTS = lefttodownload.what_is_left_to_download(input_SIFTS_files_were_found, all_SIFTS_files_from_latest_catalog)
        urls_to_target_SIFTS_files = url_formation_for_pool("SIFTS", left_to_download_SIFTS, default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        run_downloads_with_ThreadPool("SIFTS", urls_to_target_SIFTS_files, default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        return left_to_download_PDB

    if format_of_db == "PDB_assembly":
        all_data_from_catreader = latestcatreader.latest_catalog_reader()
        all_SIFTS_files_from_latest_catalog = all_data_from_catreader[2]

        download_all_PDB_assemblies = download_pdb_assemblies_list_with_lxml()
        input_PDB_assembly_files_were_found = lookfilesinside.look_what_is_inside(
            "PDB_assembly", default_input_path_to_PDB_assembly=default_input_path_to_PDB_assembly)
        try:
            len(download_all_PDB_assemblies)
        except TypeError:
            return print("Cannot reach https://ftp.wwpdb.org/pub/pdb/data/biounit/PDB/all/ maybe try again later")
        lefttodownload_PDB_assemblies = [assembly for assembly in download_all_PDB_assemblies
                                         if assembly.rsplit('/', 1)[-1] not in input_PDB_assembly_files_were_found]
        run_downloads_with_ThreadPool("PDB_assembly", lefttodownload_PDB_assemblies,
                                      default_input_path_to_PDB_assembly=default_input_path_to_PDB_assembly)

        input_SIFTS_files_were_found = lookfilesinside.look_what_is_inside("SIFTS", default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        left_to_download_SIFTS = lefttodownload.what_is_left_to_download(input_SIFTS_files_were_found, all_SIFTS_files_from_latest_catalog)
        urls_to_target_SIFTS_files = url_formation_for_pool("SIFTS", left_to_download_SIFTS, default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        run_downloads_with_ThreadPool("SIFTS", urls_to_target_SIFTS_files, default_input_path_to_SIFTS=default_input_path_to_SIFTS)
        return lefttodownload_PDB_assemblies

    if format_of_db == "all":
        all_data_from_catreader = latestcatreader.latest_catalog_reader()
        all_mmCIF_files_from_latest_catalog = all_data_from_catreader[0]
        all_PDB_files_from_latest_catalog = all_data_from_catreader[1]
        all_SIFTS_files_from_latest_catalog = all_data_from_catreader[2]

        input_mmCIF_files_were_found = lookfilesinside.look_what_is_inside("mmCIF", default_input_path_to_mmCIF=default_input_path_to_mmCIF)
        input_PDB_files_were_found = lookfilesinside.look_what_is_inside("PDB", default_input_path_to_PDB=default_input_path_to_PDB)
        input_SIFTS_files_were_found = lookfilesinside.look_what_is_inside("SIFTS", default_input_path_to_SIFTS=default_input_path_to_SIFTS)

        left_to_download_mmCIF = lefttodownload.what_is_left_to_download(input_mmCIF_files_were_found, all_mmCIF_files_from_latest_catalog)
        left_to_download_PDB = lefttodownload.what_is_left_to_download(input_PDB_files_were_found, all_PDB_files_from_latest_catalog)
        left_to_download_SIFTS = lefttodownload.what_is_left_to_download(input_SIFTS_files_were_found, all_SIFTS_files_from_latest_catalog)

        urls_to_target_mmCIF_files = url_formation_for_pool("mmCIF", left_to_download_mmCIF, default_input_path_to_mmCIF=default_input_path_to_mmCIF)
        urls_to_target_PDB_files = url_formation_for_pool("PDB", left_to_download_PDB, default_input_path_to_PDB=default_input_path_to_PDB)
        urls_to_target_SIFTS_files = url_formation_for_pool("SIFTS", left_to_download_SIFTS, default_input_path_to_SIFTS=default_input_path_to_SIFTS)

        run_downloads_with_ThreadPool("mmCIF", urls_to_target_mmCIF_files, default_input_path_to_mmCIF=default_input_path_to_mmCIF)
        run_downloads_with_ThreadPool("PDB", urls_to_target_PDB_files, default_input_path_to_PDB=default_input_path_to_PDB)
        run_downloads_with_ThreadPool("SIFTS", urls_to_target_SIFTS_files, default_input_path_to_SIFTS=default_input_path_to_SIFTS)

        # PDB_assembly
        download_all_PDB_assemblies = download_pdb_assemblies_list_with_lxml()
        input_PDB_assembly_files_were_found = lookfilesinside.look_what_is_inside("PDB_assembly")
        try:
            len(download_all_PDB_assemblies)
        except TypeError:
            return print("Cannot reach https://ftp.wwpdb.org/pub/pdb/data/biounit/PDB/all/ maybe try again later")
        lefttodownload_PDB_assemblies = [assembly for assembly in download_all_PDB_assemblies
                                         if assembly.rsplit('/', 1)[-1] not in input_PDB_assembly_files_were_found]
        run_downloads_with_ThreadPool("PDB_assembly", lefttodownload_PDB_assemblies,
                                      default_input_path_to_PDB_assembly=default_input_path_to_PDB_assembly)

        # mmCIF_assembly
        lefttodownload_mmCIF_assemblies = list()
        input_mmCIF_assembly_files_were_found = lookfilesinside.look_what_is_inside(
            "mmCIF_assembly", default_input_path_to_mmCIF_assembly=default_input_path_to_mmCIF_assembly)

        all_mmCIF_files_4char = set()
        for mmCIF_file in all_mmCIF_files_from_latest_catalog:
            all_mmCIF_files_4char.add(mmCIF_file[:4])

        input_mmCIF_assembly_files_were_found_4char = set()
        for mmCIF_assembly_file in input_mmCIF_assembly_files_were_found:
            input_mmCIF_assembly_files_were_found_4char.add(mmCIF_assembly_file[:4])

        set_difference = all_mmCIF_files_4char - input_mmCIF_assembly_files_were_found_4char
        list_difference = list(set_difference)

        for mmCIF_id in list_difference:
            lefttodownload_mmCIF_assemblies.append(mmCIF_id + ".cif.gz")

        urls_to_target_mmCIF_assembly_files = url_formation_for_pool("mmCIF_assembly", lefttodownload_mmCIF_assemblies,
                                                                     default_input_path_to_mmCIF_assembly=default_input_path_to_mmCIF_assembly)
        run_downloads_with_ThreadPool("mmCIF_assembly", urls_to_target_mmCIF_assembly_files,
                                      default_input_path_to_mmCIF_assembly=default_input_path_to_mmCIF_assembly)

        return [left_to_download_mmCIF, left_to_download_PDB, lefttodownload_mmCIF_assemblies, lefttodownload_PDB_assemblies]


"""usage of supreme_download_master"""
# supreme_download_master("mmCIF")
# supreme_download_master("PDB")



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
from tqdm import tqdm

from ..functions.coord import get_pdb_id
from ..functions.pdbaa import (
    get_pdbaa_swiss_id,
    get_pdbaa_chainid,
    get_pdbaa_method,
    get_pdbaa_pdb_code,
    get_pdbaa_prot,
    get_pdbaa_r_factor,
    get_pdbaa_resolution,
)
from ..functions.table import lst_col
from ..functions.seq import load_record_dict, get_record_desc, get_record_seq
from ..functions.col import (
    pdb_id_col,
    pdb_code_col,
    chainid_col,
    swiss_id_col,
    prot_col,
    method_col,
    resolution_col,
    r_factor_col,
    seq_col,
    len_col,
    swiss_id_col,
)
from ..functions.path import save_table


def search_pdbaa(
    pdbaa_fasta_path,
    search_lst,
    entry_table_path=None,
    min_length=None,
    min_resolution=None,
):

    pdbaa_dict = load_record_dict(pdbaa_fasta_path)

    for search in search_lst:
        if "_" in search:
            search_col = swiss_id_col
            break
        else:
            search_col = pdb_id_col
            break

    col_lst = [
        pdb_id_col,
        pdb_code_col,
        chainid_col,
        swiss_id_col,
        prot_col,
        method_col,
        resolution_col,
        r_factor_col,
        seq_col,
        len_col,
    ]

    record_dict = {}

    for col in col_lst:
        record_dict[col] = list()

    for record in tqdm(
        list(pdbaa_dict.keys()), desc="Searching pdbaa", position=0, leave=True
    ):

        record = pdbaa_dict[record]

        desc = get_record_desc(record)

        swiss_id = get_pdbaa_swiss_id(desc)

        pdb_code = get_pdbaa_pdb_code(desc)
        chainid = get_pdbaa_chainid(desc)
        pdb_id = get_pdb_id(pdb_code, chainid)

        get_seq = False

        if search_col == swiss_id_col:
            if swiss_id in search_lst:
                get_seq = True

        elif search_col == pdb_id_col:
            if pdb_id in search_lst:
                get_seq = True

        if get_seq:

            seq = get_record_seq(record)
            length = len(seq)

            get_lst = True
            if min_length is not None:
                if length <= min_length:
                    get_lst = False

            if get_lst:

                resolution = get_pdbaa_resolution(desc)

                add_dict = True
                if min_resolution is not None:
                    if resolution <= min_resolution:
                        add_dict = False

                if add_dict:

                    record_dict[pdb_id_col].append(pdb_id)
                    record_dict[pdb_code_col].append(pdb_code)
                    record_dict[chainid_col].append(chainid)
                    record_dict[method_col].append(get_pdbaa_method(desc))
                    record_dict[resolution_col].append(resolution)
                    record_dict[r_factor_col].append(get_pdbaa_r_factor(desc))
                    record_dict[swiss_id_col].append(swiss_id)
                    record_dict[prot_col].append(get_pdbaa_prot(desc))
                    record_dict[seq_col].append(seq)
                    record_dict[len_col].append(length)

    df = pd.DataFrame(record_dict)

    total_entry = len(lst_col(df, pdb_code_col, unique=True))
    total_chains = len(lst_col(df, pdb_id_col, unique=True))

    print(
        f"Your search identified {total_chains} chains from {total_entry} PDB entries!"
    )

    if entry_table_path is not None:
        save_table(entry_table_path, df)
    else:
        return df



# -*- coding: utf-8 -*-

"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import os
import pandas as pd

from ..scripts.annot_lig import annot_lig
from ..scripts.prep_dih import prep_dih
from ..scripts.prep_edia import prep_edia
from ..scripts.build_dih_table import build_dih_table
from ..scripts.build_dih_matrix import build_dih_matrix
from ..scripts.build_rmsd_matrix import build_rmsd_matrix
from ..scripts.mask_dih_data import mask_dih_data
from ..scripts.cluster_matrix import cluster_matrix
from ..scripts.classify_matrix import classify_matrix
from ..scripts.build_dist_table import build_dist_table
from ..scripts.write_pymol_script import write_pymol_script

from ..functions.cluster import build_sum_table

from ..constants.conf import (
    conf_name_dict,
    y32_name,
    y71_name,
    sw1_name,
    sw2_name,
    in_name,
    out_name,
    outlier_name,
    disorder_name,
    gtp_name,
    noise_name,
    conf_color_dict
)
from ..constants.nuc import nuc_class_dict, nuc_class_lst
from ..constants.pml import sup_resids, show_resids

from ..functions.table import get_col_most_common, lst_col, mask_equal, make_dict
from ..functions.path import (
    load_table,
    save_table,
    load_matrix,
    delete_path,
    load_table,
    load_json,
    get_file_path,
    delete_path,
    rascore_str,
    cluster_str,
)
from ..functions.file import (
    entry_table_file,
    sifts_json_file,
    edia_json_file,
    dih_json_file,
    dih_table_file,
    fit_table_file,
    pred_table_file,
    cluster_table_file,
    result_table_file,
    sum_table_file,
    dih_matrix_file,
    rmsd_matrix_file,
    rmsd_json_file,
    dih_fit_matrix_file,
    dih_pred_matrix_file,
    rmsd_fit_matrix_file,
    rmsd_pred_matrix_file,
    pymol_pml_file,
)

from ..functions.lig import lig_col_lst
from ..functions.col import (
    pdb_code_col, 
    complete_col,
    rama_col,
    pdb_id_col,
    bio_lig_col,
    nuc_class_col, 
    cluster_col, 
    core_path_col,)


def cluster_rascore(build_path, out_path=None, name_dict=None,   
                    sw1_resids="25-40", sw2_resids="56-76", 
                    g12_resid=12, v9_resid=9,
                    y32_resid=32, y71_resid=71, 
                    y32_dist=10.5, y71_dist=8.75,
                    num_cpu=1):

    if out_path is None:
        out_path = f"{os.getcwd()}/{rascore_str}_{cluster_str}"

    if name_dict is None:
        name_dict = conf_name_dict

    entry_table_path = get_file_path(entry_table_file, dir_path=build_path)
    dih_json_path = get_file_path(dih_json_file, dir_path=build_path)

    df = load_table(entry_table_path)

    df_col_lst = list(df.columns)

    for col in [y32_name, y71_name, sw1_name, sw2_name]:
        if col in df_col_lst:
            del df[col]

    if pdb_code_col in df_col_lst:
        try:
            edia_json_path = get_file_path(edia_json_file, dir_path=build_path)
            sifts_json_path = get_file_path(sifts_json_file, dir_path=build_path)

            sifts_dict = load_json(sifts_json_path)

            pdb_code_lst = lst_col(df, pdb_code_col, unique=True)
            prep_edia(
                pdb_codes=pdb_code_lst,
                edia_dir=build_path,
                sifts_dict=sifts_dict,
                edia_json_path=edia_json_path,
                num_cpu=num_cpu,
            )
        except:
            delete_path(edia_json_path)

    dih_dict = load_json(dih_json_path)
    edia_dict = load_json(edia_json_path)

    if dih_dict is None:
        prep_dih(lst_col(df,core_path_col),dih_json_path=dih_json_path,num_cpu=num_cpu)
        dih_dict = load_json(dih_json_path)

    if len([x for x in lig_col_lst if x not in df_col_lst]) > 0:
        df = annot_lig(
            df=df,
            num_cpu=num_cpu,
        )

    if nuc_class_col not in df_col_lst:
        df[nuc_class_col] = df[bio_lig_col].map(nuc_class_dict).fillna(gtp_name)

    dist_df = build_dist_table(
            df,
            x_resids=[y32_resid, y71_resid],
            y_resids=[g12_resid, v9_resid],
            x_atomids=['OH','OH'],
            y_atomids=['CA','CA'],
            atom_dist_col_lst=[y32_name, y71_name]
        )

    dist_dict = {y32_name: y32_dist, y71_name: y71_dist}

    for index in list(dist_df.index.values):

        nuc_class = dist_df.at[index, nuc_class_col]

        for resid_name in [y32_name, y71_name]:

            atom_dist = dist_df.at[index, resid_name]

            if atom_dist == 999.00:
                group_name = disorder_name
            else:
                group_name = resid_name
                if atom_dist <= dist_dict[resid_name]:
                    group_name += in_name
                else:
                    group_name += out_name

            dist_df.at[index, resid_name] = group_name


    for resid_name in [y32_name, y71_name]:

        df[resid_name] = (
            df[pdb_id_col]
            .map(
                make_dict(lst_col(dist_df, pdb_id_col), lst_col(dist_df, resid_name))
            )
        )

    for resid_name in [y32_name, y71_name]:

        if resid_name == y32_name:
            loop_name = sw1_name
            loop_resids = sw1_resids
        elif resid_name == y71_name:
            loop_name = sw2_name
            loop_resids = sw2_resids

        loop_result_df = pd.DataFrame()
        loop_sum_df = pd.DataFrame()

        loop_path = f"{out_path}/{loop_name}"

        total_complete = 0
        total_clustered = 0
        
        for group_name in lst_col(df, resid_name, unique=True):

            group_df = mask_equal(df, resid_name, group_name)

            for nuc_class in nuc_class_lst:

                group_nuc_df = mask_equal(group_df, nuc_class_col, nuc_class)

                group_nuc_name = f"{group_name}.{nuc_class}"

                dih_table_path = get_file_path(
                    dih_table_file, dir_str=group_nuc_name, dir_path=loop_path
                )

                fit_table_path = get_file_path(
                    fit_table_file, dir_str=group_nuc_name, dir_path=loop_path
                )

                pred_table_path = get_file_path(
                    pred_table_file, dir_str=group_nuc_name, dir_path=loop_path
                )

                cluster_table_path = get_file_path(
                    cluster_table_file, dir_str=group_nuc_name, dir_path=loop_path
                )

                result_table_path = get_file_path(
                    result_table_file, dir_str=group_nuc_name, dir_path=loop_path
                )

                dih_matrix_path = get_file_path(
                    dih_matrix_file, dir_str=group_nuc_name, dir_path=loop_path
                )

                rmsd_matrix_path = get_file_path(
                    rmsd_matrix_file, dir_str=group_nuc_name, dir_path=loop_path
                )

                rmsd_json_path = get_file_path(
                    rmsd_json_file, dir_str=group_nuc_name, dir_path=loop_path
                )

                dih_fit_matrix_path = get_file_path(
                    dih_fit_matrix_file, dir_str=group_nuc_name, dir_path=loop_path
                )

                dih_pred_matrix_path = get_file_path(
                    dih_pred_matrix_file, dir_str=group_nuc_name, dir_path=loop_path
                )

                rmsd_fit_matrix_path = get_file_path(
                    rmsd_fit_matrix_file, dir_str=group_nuc_name, dir_path=loop_path
                )

                rmsd_pred_matrix_path = get_file_path(
                    rmsd_pred_matrix_file, dir_str=group_nuc_name, dir_path=loop_path
                )

                build_dih_table(
                    df=group_nuc_df,
                    dih_dict=dih_dict,
                    dih_table_path=dih_table_path,
                    bb_resids=loop_resids,
                )
                dih_df = load_table(dih_table_path)

                if len(dih_df) > 15 and disorder_name not in group_nuc_name:

                    rmsd_dict = load_json(rmsd_json_path)
                    build_rmsd_matrix(
                        fit_df=dih_df,
                        rmsd_matrix_path=rmsd_matrix_path,
                        sup_resids=sup_resids,
                        rmsd_resids=loop_resids,
                        rmsd_atomids="CA",
                        pair_aln=False,
                        rmsd_dict=rmsd_dict,
                        rmsd_json_path=rmsd_json_path,
                    )

                    build_dih_matrix(
                        fit_df=dih_df,
                        max_norm_path=dih_matrix_path,
                    )

                    dih_matrix = load_matrix(dih_matrix_path)
                    rmsd_matrix = load_matrix(rmsd_matrix_path)

                    mask_dih_data(
                        df=dih_df,
                        matrix=dih_matrix,
                        fit_table_path=fit_table_path,
                        fit_matrix_path=dih_fit_matrix_path,
                        pred_table_path=pred_table_path,
                        pred_matrix_path=dih_pred_matrix_path,
                        edia_dict=edia_dict,
                        edia_min=0.4,
                        edia_atomids="O",
                    )
                    mask_dih_data(
                        df=dih_df,
                        matrix=rmsd_matrix,
                        fit_table_path=fit_table_path,
                        fit_matrix_path=rmsd_fit_matrix_path,
                        pred_table_path=pred_table_path,
                        pred_matrix_path=rmsd_pred_matrix_path,
                        edia_dict=edia_dict,
                        edia_min=0.4,
                        edia_atomids="O",
                    )

                    fit_df = load_table(fit_table_path)
                    dih_fit_matrix = load_matrix(dih_fit_matrix_path)
                    rmsd_fit_matrix = load_matrix(rmsd_fit_matrix_path)

                    cluster_matrix(
                        df=fit_df,
                        matrix=dih_fit_matrix,
                        cluster_table_path=cluster_table_path,
                        max_nn_dist=0.45,
                        constr_matrix=rmsd_fit_matrix,
                        max_constr_dist=1.2,
                        merge_constr_dist=1.2,
                        min_samples_range="7-15",
                        min_min_samples=7,
                        min_pdb=5,
                    )

                    cluster_df = load_table(cluster_table_path)

                    pred_df = load_table(pred_table_path)

                    dih_pred_matrix = load_matrix(dih_pred_matrix_path)
                    rmsd_pred_matrix = load_matrix(rmsd_pred_matrix_path)

                    classify_matrix(
                        cluster_df=cluster_df,
                        pred_df=pred_df,
                        fit_matrix=dih_fit_matrix,
                        pred_matrix=dih_pred_matrix,
                        result_table_path=result_table_path,
                        fit_constr_matrix=rmsd_fit_matrix,
                        pred_constr_matrix=rmsd_pred_matrix,
                        max_nn_dist=0.45,
                        max_constr_dist=1.2,
                    )
                else:
                    cluster_df = dih_df.copy(deep=True)
                    cluster_df[cluster_col] = noise_name

                    result_df = cluster_df.copy(deep=True)

                    save_table(cluster_table_path, cluster_df)
                    save_table(result_table_path, result_df)

                cluster_df = load_table(cluster_table_path)
                result_df = load_table(result_table_path)

                rename_dict = dict()

                i = 1
                for cluster in lst_col(result_df,cluster_col,unique=True):
                    name = cluster
                    if cluster != noise_name:
                        rama = [x for x in get_col_most_common(mask_equal(result_df,cluster_col,cluster), rama_col) if "-" not in x][0]
                        name = f"{group_nuc_name}-Unknown-{i}"
                        i += 1
                        if group_nuc_name in list(name_dict.keys()):
                            if rama in list(name_dict[group_nuc_name].keys()):
                                name = name_dict[group_nuc_name][rama]
                                i -= 1                        

                    rename_dict[cluster] = name

                result_df[cluster_col] = result_df[cluster_col].map(rename_dict)

                cluster_df[cluster_col] = cluster_df[pdb_id_col].map(
                    make_dict(
                        lst_col(result_df, pdb_id_col), lst_col(result_df, cluster_col)
                    )
                )

                save_table(cluster_table_path, cluster_df)

                for index in list(result_df.index.values):
                    cluster = result_df.at[index, cluster_col]
                    if noise_name in cluster:
                        complete =  str(result_df.at[index, complete_col])
                        if complete == str(True):
                            cluster = outlier_name
                        elif complete == str(False):
                            cluster = disorder_name
                        result_df.at[index,cluster_col] = cluster

                loop_result_df = pd.concat([loop_result_df, result_df], sort=False)

                total_complete += len(cluster_df)
                total_clustered += len(mask_equal(cluster_df, cluster_col, [x for x in lst_col(cluster_df, cluster_col) if noise_name not in x]))

        loop_result_table_path = get_file_path(
            result_table_file, dir_str=loop_name, dir_path=out_path
        )
        loop_sum_table_path = get_file_path(
            sum_table_file, dir_str=loop_name, dir_path=out_path
        )

        loop_sum_df = build_sum_table(loop_result_df)

        save_table(loop_result_table_path, loop_result_df, fillna="-")
        save_table(loop_sum_table_path, loop_sum_df, fillna="-")

        cluster_dict = make_dict(
            lst_col(loop_result_df, pdb_id_col), lst_col(loop_result_df, cluster_col)
        )

        df[loop_name] = df[pdb_id_col].map(cluster_dict)

        final_spatial = len(mask_equal(loop_result_df, resid_name, [x for x in lst_col(loop_result_df, resid_name) if disorder_name not in x]))
        final_conformation = len(mask_equal(loop_result_df, cluster_col, [x for x in lst_col(loop_result_df, cluster_col) if outlier_name not in x and disorder_name not in x]))

        print(f"Results for {loop_name}: Total Well Modeled = {total_complete}; Total Clustered = {total_clustered}; Final with Spatial Labels = {final_spatial}; Final with Conformational Labels = {final_conformation}")

    save_table(entry_table_path, df)

    for loop_name in [sw1_name, sw2_name]:

        if loop_name == sw1_name:
            loop_resids = sw1_resids
        elif loop_name == sw2_name:
            loop_resids = sw2_resids

        pymol_pml_path = get_file_path(
            f"{loop_name}_{pymol_pml_file}", dir_path=out_path
        )

        if loop_name == sw1_name:
            stick_resids = [32]
        elif loop_name == sw2_name:
            stick_resids = [71]

        write_pymol_script(
            mask_equal(df, loop_name, [x for x in lst_col(df, loop_name, unique=True) if noise_name not in x]),
            pymol_pml_path,
            group_col=loop_name,
            stick_resids=stick_resids,
            loop_resids=loop_resids,
            style_ribbon=True,
            thick_bb=False,
            show_bio=True,
            sup_group=True,
            color_palette=conf_color_dict[loop_name],
            sup_resids=sup_resids,
            show_resids=show_resids,
        )

    print("Rascore clustering complete!")


# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import pandas as pd
import streamlit as st
from random import randint

from ..constants.conf import (sw1_color, sw2_color, sw1_name, sw2_name, 
                            loop_resid_dict, sw1_resids, sw2_resids)
from ..constants.prot import prot_color_dict
from ..constants.pharm import pharm_color_dict
from ..constants.pml import sup_resids, show_resids
from ..scripts.write_pymol_script import write_pymol_script
from ..functions.table import extract_int, lst_col, str_to_dict
from ..functions.lst import str_to_lst, lst_nums, res_to_lst
from ..functions.gui import (
    load_st_table,
    show_st_table,
    mask_st_table,
    show_st_structure,
    write_st_end,
    get_html_text,
    download_st_file,
    get_neighbor_path,
    ribbon_name,
    trace_name,
    standard_name,
    aa_name,
)
from ..functions.download import download_unzip
from ..functions.file import pymol_pml_file
from ..functions.path import pages_str, data_str, delete_path, get_file_path, path_exists
from ..functions.col import (
    rename_col_dict,
    pdb_code_col,
    chainid_col,
    bio_lig_col,
    ion_lig_col,
    pharm_lig_col,
    chem_lig_col,
    mod_lig_col,
    mem_lig_col,
    gene_class_col,
    method_col,
    resolution_col,
    r_factor_col,
    space_col,
    mut_status_col,
    nuc_class_col,
    prot_class_col,
    match_class_col,
    pocket_class_col,
    interf_class_col,
    bound_prot_col,
    bound_prot_swiss_id_col,
    bound_prot_chainid_col,
    bound_lig_cont_col,
    bound_prot_cont_col,
    pharm_class_col,
    sw1_col,
    sw2_col,
    y32_col,
    y71_col,
    date_col,
    core_path_col
)

sw1_resid_lst = res_to_lst(loop_resid_dict[sw1_name])
sw2_resid_lst = res_to_lst(loop_resid_dict[sw2_name])

def pdb_page():

    st.markdown("# Search PDB")

    st.markdown("---")

    df = load_st_table(__file__)

    st.sidebar.markdown("### PDB Selection")

    pdb_code = st.sidebar.selectbox(
        "Entry", [x.upper() for x in lst_col(df, pdb_code_col, unique=True)]
    )
    pdb_df = mask_st_table(df, {pdb_code_col: pdb_code.lower()})

    chainid = st.sidebar.selectbox("Chain", lst_col(pdb_df, chainid_col))
    chainid_df = mask_st_table(pdb_df, {chainid_col: chainid})

    gene_class = chainid_df.at[0, gene_class_col]
    mut_status = chainid_df.at[0, mut_status_col]

    st.markdown(f"#### PDB: [{pdb_code}](https://www.rcsb.org/structure/{pdb_code}) (Chain {chainid}) - {gene_class}({mut_status})")

    left_col, right_col = st.columns(2)

    left_col.markdown("##### General Information")
    for col in [method_col, resolution_col, r_factor_col, space_col, date_col]:
        left_col.markdown(f"**{rename_col_dict[col]}:** {chainid_df.at[0,col]}")

    right_col.markdown("##### Molecular Annotations")

    annot_df = pd.DataFrame()
    for i, col in enumerate(
        [
            nuc_class_col,
            prot_class_col,
            pocket_class_col,
            match_class_col,
            interf_class_col,
        ]
    ):
        annot_df.at[i, "Molecular Content"] = rename_col_dict[col]
        annot_df.at[i, "Annotation"] = chainid_df.at[0, col]

    show_st_table(annot_df, st_col=right_col)

    if chainid_df.at[0,bound_prot_col] != "None":
        for col in [bound_prot_col, bound_prot_swiss_id_col]:
            left_col.markdown(f"**{rename_col_dict[col]}:** {chainid_df.at[0,col]}")

    st_col_lst = st.columns(4)

    for i, col in enumerate([sw1_col, sw2_col, y32_col, y71_col]):

        if col in [sw1_col, y32_col]:
            color = sw1_color
        elif col in [sw2_col, y71_col]:
            color = sw2_color

        st_col_lst[i].markdown(f"##### {rename_col_dict[col]}")
        st_col_lst[i].markdown(get_html_text({chainid_df.at[0, col]:color}, 
        font_size="large"),unsafe_allow_html=True)

    st.markdown("---")

    left_check_col, middle_check_col, right_check_col = st.columns(3)
    left_view_col, right_view_col = st.columns([0.4,0.6])

    left_view_col.markdown("##### Viewer Settings")

    stick_resids = [32, 71]
    zoom_resids = None
    cartoon_trans = 1.0
    surface_trans = 0.0
    zoom = 1.5

    left_check_col.markdown("##### Bound Ligands")

    lig_check_dict = dict()
    for col in [
        bio_lig_col,
        ion_lig_col,
        pharm_lig_col,
        chem_lig_col,
        mod_lig_col,
        mem_lig_col,
    ]:
        lig_lst = str_to_lst(chainid_df.at[0, col])
        if "None" not in lig_lst:
            for lig in lig_lst:
                lig_check_dict[lig] = left_check_col.checkbox(f"{rename_col_dict[col]}: {lig}")

    if len(lig_check_dict.keys()) == 0:
        left_check_col.write("No bound ligands.")

    pharm_lig_lst = str_to_lst(chainid_df[pharm_lig_col].iloc[0]) 

    pharm_on = False
    if "None" not in pharm_lig_lst:
        if left_view_col.checkbox("Display Inhibitor Site"):

            pharm_on = True

            pharm_lig = left_view_col.selectbox("Select Inhibitor Site", pharm_lig_lst)
            pharm_cont_dict = str_to_dict(chainid_df[bound_lig_cont_col].iloc[0], return_int=True)

            stick_resids = pharm_cont_dict[pharm_lig]
            zoom_resids = {"resn": pharm_lig}
            cartoon_trans = 0.5
            surface_trans = 0.7
            zoom = 1.0

    middle_check_col.markdown("##### Bound Proteins")

    prot_check_dict = dict()
    for prot_lst in str_to_lst(chainid_df.at[0, bound_prot_chainid_col]):
        if "None" not in prot_lst:
            for prot in prot_lst:
                prot_check_dict[prot] = middle_check_col.checkbox(f"Chain {prot}")

    if len(prot_check_dict.keys()) == 0:
        middle_check_col.write("No bound proteins.")


    bound_prot_lst = str_to_lst(chainid_df[bound_prot_chainid_col].iloc[0]) 

    if "None" not in bound_prot_lst:
        if left_view_col.checkbox("Display Bound Protein Site"):

            if pharm_on:
                left_view_col.warning("Cannot display bound protein and inhibitor sites simultaneously.")
            else:
                bound_prot = left_view_col.selectbox("Select Inhibitor Site", bound_prot_lst)
                prot_cont_dict = str_to_dict(chainid_df[bound_prot_cont_col].iloc[0], return_int=True)

                stick_resids = prot_cont_dict[bound_prot]
                cartoon_trans = 0.5
                surface_trans = 0.0
                zoom = 1.0

    right_check_col.markdown("##### Mutation Sites")

    mut_check_dict = dict()
    for mut in str_to_lst(chainid_df.at[0, mut_status_col]):
        if mut != "WT":
            mut_check_dict[extract_int(mut)] = right_check_col.checkbox(mut)

    if len(mut_check_dict.keys()) == 0:
        right_check_col.write("Not mutated.")

    stick_resids = left_view_col.multiselect("Displayed Residues", lst_nums(1, 189), default=stick_resids)

    label_resids = left_view_col.checkbox("Label Residues", value=True)

    scheme_dict = {standard_name: False, aa_name: True}
    aa_scheme = scheme_dict[left_view_col.radio("Color Scheme", [standard_name, aa_name])]

    style_dict = {ribbon_name: "oval", trace_name: "trace"}
    pymol_style_dict = {ribbon_name: False, trace_name: True}

    bb_style = left_view_col.radio("Cartoon Style", [ribbon_name, trace_name])
    cartoon_style = style_dict[bb_style]
    pymol_cartoon_style = pymol_style_dict[bb_style]

    cartoon_trans = left_view_col.slider(
        "Cartoon Transparency", min_value=0.0, max_value=1.0, value=cartoon_trans
    )

    surface_trans = left_view_col.slider(
        "Surface Transparency", min_value=0.0, max_value=1.0, value=surface_trans
    )

    spin_on = left_view_col.checkbox("Rotate Structure")

    all_chains = False
    if len(pdb_df) > 1:
        all_chains = left_view_col.checkbox("Show All Chains")


    show_st_structure(chainid_df,
                        mut_resids=list(mut_check_dict.keys()), stick_resids=stick_resids,
                        label_muts=mut_check_dict, label_resids=label_resids, 
                        label_ligs=lig_check_dict, label_prots=prot_check_dict, 
                        cartoon_style=cartoon_style,
                        cartoon_trans=cartoon_trans, 
                        surface_trans=surface_trans,
                        mut_trans=surface_trans,
                        aa_scheme=aa_scheme,
                        spin_on=spin_on,
                        all_chains=all_chains,
                        zoom_resids=zoom_resids,
                        zoom=zoom,
                        width=500,
                        height=500,
                        st_col=right_view_col)

    st.markdown("---")

    left_get_col, right_get_col = st.columns(2)

    pdb_name = "PDB"
    cif_name = "mmCIF"
    rcsb_name = "RCSB PDB"
    renum_name = "PDBrenum"

    data_dir = get_neighbor_path(__file__, pages_str, data_str)

    left_get_col.markdown("##### Download Coordinate File")
    right_get_col.markdown("##### Download PyMOL Script")

    format_dict = {cif_name:"cif", pdb_name: "pdb"}

    file_format = left_get_col.radio("File Format", [cif_name, pdb_name])

    file_source = left_get_col.radio("File Source", [renum_name, rcsb_name])

    file_str = pdb_code
    if file_source == renum_name:
        file_str += "_renum"
    file_str += "."
    file_str += format_dict[file_format]
    
    coord_file_name = left_get_col.text_input(
        label="Coordinate File Name",
        value=file_str,
    )

    coord_file_path = get_file_path(
        f"{coord_file_name}_{randint(0,3261994)}",
        dir_path=data_dir,
    )

    if left_get_col.button("Prepare Coordinate File"):
        with st.spinner(text="Preparing Coordinate File"):
            if file_source == rcsb_name:
                coord_url = f"https://files.rcsb.org/download/{pdb_code.lower()}.{format_dict[file_format]}.gz"
            elif file_source == renum_name:
                coord_url = f"http://dunbrack3.fccc.edu/PDBrenum/output_{file_format}/{pdb_code.lower()}_renum.{format_dict[file_format]}.gz"
               
            download_unzip(coord_url, coord_file_path)
    
        download_st_file(
            coord_file_path,
            coord_file_name,
            f"Download Coordinate File",
            st_col=left_get_col
        )

        delete_path(coord_file_path)


    coord_path_col = pdb_code_col
    if len(
        [x for x in lst_col(chainid_df, core_path_col) if path_exists(x)]
    ) == len(chainid_df):
        coord_path_col = core_path_col
    
    pymol_file_name = right_get_col.text_input(
        label="PyMOL Script Name",
        value=f"{pdb_code}_{chainid}_{pymol_pml_file}",
    )

    fetch_path = right_get_col.text_input(
                label="Fetch Path (e.g., /Users/mitch-parker/rascore)",
            )

    pymol_file_path = get_file_path(
        f"{pymol_pml_file}_{randint(0,3261994)}",
        dir_path=data_dir,
    )

    if right_get_col.button("Create PyMOL Script"):
        with st.spinner(text="Creating PyMOL Script"):
            write_pymol_script(
                chainid_df,
                pymol_file_path,
                stick_resids=stick_resids,
                loop_resids=[sw1_resids, sw2_resids],
                style_ribbon=pymol_cartoon_style,
                thick_bb=False,
                color_group=False,
                color_palette=[sw1_color, sw2_color],
                show_bio=True,
                show_ion=True,
                show_pharm=pharm_color_dict[chainid_df[pharm_class_col].iloc[0]],
                show_chem=True,
                show_mod=True,
                show_mem=True,
                show_prot=prot_color_dict[chainid_df[prot_class_col].iloc[0]],
                sup_resids=sup_resids,
                show_resids=show_resids,
                cartoon_transp=1-cartoon_trans,
                surface_transp=1-surface_trans,
                coord_path_col=coord_path_col,
                fetch_path=fetch_path,
            )

        download_st_file(
            pymol_file_path,
            pymol_file_name,
            f"Download PyMOL Script",
            st_col=right_get_col
        )

        delete_path(pymol_file_path)

    write_st_end()


# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
import streamlit as st
from random import randint

from ..scripts.write_pymol_script import write_pymol_script

from ..functions.gui import (
    load_st_table,
    rename_st_cols,
    mask_st_table,
    show_st_table,
    download_st_df,
    download_st_file,
    show_st_dataframe,
    write_st_end,
    reorder_st_cols
)
from ..functions.lst import lst_nums
from ..constants.nuc import nuc_class_lst
from ..constants.gene import gene_class_lst
from ..functions.table import lst_col, fix_col, mask_equal, make_dict
from ..functions.color import get_lst_colors, get_hex
from ..constants.pml import sup_resids, show_resids, sup_pdb_code, sup_chainid, mono_view
from ..functions.col import (
    rename_col_dict,
    pdb_id_col,
    sw1_col,
    sw2_col,
    y32_col,
    y71_col,
    mut_status_col,
    prot_class_col,
    match_class_col,
    pocket_class_col,
    interf_class_col,
    gene_class_col,
    nuc_class_col,
    pdb_code_col,
    modelid_col,
    core_path_col,
    bio_lig_col,
    ion_lig_col,
    pharm_lig_col,
    chem_lig_col,
    mod_lig_col,
    mem_lig_col,
    pocket_lig_col,
    bound_prot_chainid_col,
)
from ..functions.path import (
    get_file_path,
    get_neighbor_path,
    delete_path,
    path_exists,
    pages_str,
    data_str,
    rascore_str,
)

from ..constants.conf import y32_name, y71_name, sw1_name, sw2_name, loop_resid_dict, conf_color_dict, sw1_color,sw2_color
from ..functions.file import (
    entry_table_file,
    sum_table_file,
    pymol_pml_file,
)
from ..scripts.write_pymol_script import pymol_color_dict

reverse_col_dict = make_dict(list(rename_col_dict.values()),list(rename_col_dict.keys()))


def query_page():

    st.markdown("## Query Database")

    st.markdown("---")

    df = load_st_table(__file__)

    conf_col_lst = [sw1_col, sw2_col, y32_col, y71_col]

    annot_col_lst = [
        mut_status_col,
        prot_class_col,
        pocket_class_col,
        match_class_col,
        interf_class_col,
    ]

    mask_dict = dict()

    mask_df = df.copy(deep=True)

    st.sidebar.markdown(
        "**Note.** Selections dynamically update from top to bottom. Multiple selections are possible."
    )

    st.sidebar.markdown("## Conformation Selection")

    for col in conf_col_lst:

        mask_dict[col] = st.sidebar.multiselect(
            rename_col_dict[col],
            lst_col(mask_df, col, unique=True),
        )

        mask_df = mask_st_table(mask_df, mask_dict)

    st.sidebar.markdown("## Annotation Selection")

    for col in annot_col_lst:

        mask_dict[col] = st.sidebar.multiselect(
            rename_col_dict[col],
            lst_col(mask_df, col, unique=True),
        )

        mask_df = mask_st_table(mask_df, mask_dict)

    if len(mask_df) == 0:
        st.warning("No Structures Available Based On Selected Query")
    else:

        left_query_col, right_query_col = st.columns(2)

        gene_lst = lst_col(mask_df, gene_class_col)

        gene_radio_lst = [f"All (N={len(gene_lst)})"]
        for gene in gene_class_lst:
            total_gene = len([x for x in gene_lst if x == gene])
            if total_gene > 0:
                gene_radio_lst.append(f"{gene} (N={total_gene})")

        gene_class = left_query_col.radio(
            f"{rename_col_dict[gene_class_col]}", gene_radio_lst
        )

        gene_df = mask_st_table(mask_df, {gene_class_col: gene_class.split(" (")[0]})

        nuc_lst = lst_col(gene_df, nuc_class_col)

        nuc_radio_lst = [f"All (N={len(nuc_lst)})"]
        for nuc in nuc_class_lst:
            total_nuc = len([x for x in nuc_lst if x == nuc])
            if total_nuc > 0:
                nuc_radio_lst.append(f"{nuc} (N={total_nuc})")

        nuc_class = right_query_col.radio(
            f"{rename_col_dict[nuc_class_col]}", nuc_radio_lst
        )

        gene_nuc_df = mask_st_table(gene_df, {nuc_class_col: nuc_class.split(" (")[0]})

        left_conf_col, right_conf_col = st.columns(2)

        for table_col in [y32_name, y71_name, sw1_name, sw2_name]:

            if table_col in [y32_name, sw1_name]:
                conf_col = left_conf_col
            elif table_col in [y71_name, sw2_name]:
                conf_col = right_conf_col

            conf_col.markdown(f"##### {rename_col_dict[table_col]}")

            loop_df = (
                pd.pivot_table(
                    data=rename_st_cols(gene_nuc_df),
                    index=[rename_col_dict[table_col]],
                    columns=rename_col_dict[gene_class_col],
                    values=rename_col_dict[pdb_id_col],
                    aggfunc="nunique",
                    margins=True,
                )
                .fillna("")
            )

            for col in list(loop_df.columns):
                loop_df[col] = loop_df[col].map(str)
                loop_df = fix_col(loop_df, col)

            loop_df = reorder_st_cols(loop_df, table_col, gene_class_col)

            loop_df = loop_df.reset_index()

            show_st_table(loop_df, st_col=conf_col)

        with st.expander("PyMOL Script", expanded=False):

            st.markdown("#### PyMOL Script")

            pymol_lst = [x for x in list(pymol_color_dict.keys()) if x != pocket_lig_col]

            coord_path_col = pdb_code_col
            sup_coord_path = sup_pdb_code
            if len(
                [x for x in lst_col(gene_nuc_df, core_path_col) if path_exists(x)]
            ) == len(gene_nuc_df):
                coord_path_col = core_path_col
                sup_coord_path = lst_col(
                    mask_equal(df, pdb_id_col, f"{sup_pdb_code}{sup_chainid}"),
                    core_path_col,
                )[0]

            group_col = st.selectbox("Group By Selection", [rename_col_dict[x] for x in conf_col_lst + annot_col_lst])

            show_lst = st.multiselect(
                "Show Molecular Contents",
                [rename_col_dict[x] for x in pymol_lst],default=[rename_col_dict[bio_lig_col],rename_col_dict[pharm_lig_col]]
            )

            left_pymol_col, right_pymol_col = st.columns(2)

            show_color_dict = dict()
            for col in list(pymol_color_dict.keys()):
                rename_col = rename_col_dict[col]
                color = None
                if rename_col in show_lst:
                    color = left_pymol_col.color_picker(f"{rename_col} Color", get_hex(pymol_color_dict[col]))

                show_color_dict[col] = color

            style_dict = {"Ribbon": False, "Trace": True}

            style_ribbon = style_dict[
                right_pymol_col.radio("Cartoon Style", ["Ribbon", "Trace"])
            ]


            stick_resids = right_pymol_col.multiselect(
                'Stick Residues',
                lst_nums(1,189),
                default=[32,71]
            )

            loop_resids = list()
            for loop_name in list(loop_resid_dict.keys()):
                if right_pymol_col.checkbox(f"Highlight {loop_name}",value=True):
                    loop_resids.append(loop_resid_dict[loop_name])
            if len(loop_resids) == 0:
                loop_resids = [show_resids]
                

            color_group = left_pymol_col.checkbox("Color Loops By Group")
            left_pymol_col.write("**Default:** Colors SW1 pink and SW2 purple")

            if color_group:
                if group_col == rename_col_dict[sw1_name]: 
                    color_palette = conf_color_dict[sw1_name]
                elif group_col == rename_col_dict[sw2_name]: 
                    color_palette = conf_color_dict[sw2_name]
                else:
                    group_color_palette = get_lst_colors(lst_col(rename_st_cols(gene_nuc_df),group_col,unique=True),return_dict=True)
                    if len(group_color_palette.items()) > 10:
                        color_palette = group_color_palette.copy()
                    else:
                        left_color_col, right_color_col = st.columns(2)
                        color_palette = dict()
                        i = 0
                        for group, color in group_color_palette.items():
                            if i == 0:
                                color_palette[group] = left_color_col.color_picker(f"{group} Color", get_hex(color))
                                i += 1
                            elif i == 1:
                                color_palette[group] = right_color_col.color_picker(f"{group} Color", get_hex(color))
                                i = 0
            else:
                color_palette = [sw1_color,sw2_color]

            fetch_path = st.text_input(
                label="Fetch Path (e.g., /Users/mitch-parker/rascore)",
            )

            pymol_file_path = get_file_path(
                f"{pymol_pml_file}_{randint(0,3261994)}",
                dir_path=get_neighbor_path(__file__, pages_str, data_str),
            )

            pymol_file_name = st.text_input(
                label="PyMOL Script Name",
                value=f"{rascore_str}_{pymol_pml_file}",
            )
            if st.button("Create PyMOL Script"):
                with st.spinner(text="Creating PyMOL Script"):
                    write_pymol_script(
                        gene_nuc_df,
                        pymol_file_path,
                        group_col=[x for x in conf_col_lst + annot_col_lst if rename_col_dict[x] == group_col][0],
                        stick_resids=stick_resids,
                        loop_resids=loop_resids,
                        style_ribbon=style_ribbon,
                        thick_bb=False,
                        color_group=color_group,
                        color_palette=color_palette,
                        show_bio=show_color_dict[bio_lig_col],
                        show_ion=show_color_dict[ion_lig_col],
                        show_pharm=show_color_dict[pharm_lig_col],
                        show_chem=show_color_dict[chem_lig_col],
                        show_mod=show_color_dict[mod_lig_col],
                        show_mem=show_color_dict[mem_lig_col],
                        show_pocket=show_color_dict[pocket_lig_col],
                        show_prot=show_color_dict[bound_prot_chainid_col],
                        sup_resids=sup_resids,
                        show_resids=show_resids,
                        sup_coord_path=sup_coord_path,
                        sup_chainid=sup_chainid,
                        set_view=mono_view,
                        coord_path_col=coord_path_col,
                        fetch_path=fetch_path,
                    )

                download_st_file(
                    pymol_file_path,
                    pymol_file_name,
                    f"Download PyMOL Script",
                )

                delete_path(pymol_file_path)

        with st.expander("Summary Table", expanded=False):

            st.markdown("#### Summary Table")

            left_sum_col, right_sum_col = st.columns(2)

            sum_col_lst = [gene_class_col, nuc_class_col] + conf_col_lst + annot_col_lst

            row_lst = left_sum_col.multiselect(
                "Rows",
                [rename_col_dict[x] for x in sum_col_lst],
                rename_col_dict[sw2_col],
            )

            col_lst = right_sum_col.multiselect(
                "Columns",
                [rename_col_dict[x] for x in sum_col_lst],
                default=rename_col_dict[sw1_col],
            )

            gene_nuc_df = rename_st_cols(gene_nuc_df)

            if len([x for x in row_lst if x in col_lst]) > 0:
                st.warning("Cannot Have Same Selection in Rows and Columns")
            else:
                if len(row_lst) > 0 and len(col_lst) > 0:
                    sum_df = (
                        pd.pivot_table(
                            data=gene_nuc_df,
                            index=row_lst,
                            columns=col_lst,
                            values=rename_col_dict[pdb_id_col],
                            aggfunc="nunique",
                            margins=True,
                        )
                        .fillna("")
                    )

                    for col in list(sum_df.columns):
                        sum_df[col] = sum_df[col].map(str)
                        sum_df = fix_col(sum_df, col)

                
                    if len(row_lst) == 1 and len(col_lst) == 1:
                        sum_df = reorder_st_cols(sum_df, reverse_col_dict[row_lst[0]], reverse_col_dict[col_lst[0]])

                    sum_df = sum_df.reset_index()

                    show_st_table(sum_df)

                    sum_file_name = st.text_input(
                        label="Summary File Name",
                        value=f"{rascore_str}_{sum_table_file}",
                    )

                    download_st_df(sum_df, sum_file_name, "Download Summary Table")

        with st.expander("Entries Table", expanded=False):

            st.markdown("#### Entries Table")

            del gene_nuc_df[rename_col_dict[pdb_id_col]]
            del gene_nuc_df[rename_col_dict[modelid_col]]

            show_st_dataframe(gene_nuc_df)

            entries_file_name = st.text_input(
                label="Entries File Name",
                value=f"{rascore_str}_{entry_table_file}",
            )
            download_st_df(gene_nuc_df, entries_file_name, "Download Entries Table")

        write_st_end()



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import os
import pandas as pd
from tqdm import tqdm

from ..scripts.annot_lig import annot_lig
from ..scripts.prep_dih import prep_dih
from ..scripts.build_dih_matrix import build_dih_matrix
from ..scripts.build_dih_table import build_dih_table
from ..scripts.build_dist_table import build_dist_table
from ..scripts.classify_matrix import classify_matrix
from ..scripts.write_pymol_script import write_pymol_script

from ..functions.table import (
    mask_equal,
    lst_to_str,
    lst_col,
    type_lst,
    merge_tables,
    make_dict,
    fix_col,
    core_path_col,
    chainid_col,
    modelid_col,
    pdb_id_col,
)
from ..functions.path import (
    load_matrix,
    save_table,
    get_file_name,
    get_file_path,
    load_lst,
    load_table,
    get_neighbor_path,
    rascore_str,
    classify_str,
    cluster_str,
    pipelines_str,
    data_str,
)
from ..functions.cluster import build_sum_table
from ..functions.coord import load_coord, get_modelid, get_chainid
from ..functions.lig import lig_col_lst
from ..functions.col import (
    id_col,
    nuc_class_col,
    bio_lig_col,
    cluster_col,
    hb_status_col,
    complete_col,
)
from ..functions.file import (
    cluster_table_file,
    dih_fit_matrix_file,
    result_table_file,
    sum_table_file,
    pymol_pml_file,
    pred_matrix_file,
)
from ..constants.nuc import (
    gtp_name,
    nuc_class_dict,
    nuc_class_dict,
    gtp_atomids,
    nuc_class_lst,
)
from ..constants.pharm import (
    pharm_site_dict,    
)
from ..constants.conf import (
    y32_name,
    y71_name,
    sw1_name,
    sw2_name,
    in_name,
    out_name,
    noise_name,
    outlier_name,
    disorder_name,
    conf_color_dict
)
from ..constants.pml import sup_resids, show_resids


def classify_rascore(file_paths, out_path=None, dih_dict=None, 
                    sw1_resids="25-40", sw2_resids="56-76", 
                    y32_resid=32, y71_resid=71, 
                    g12_resid=12, v9_resid=9,
                    y32_dist=10.5, y71_dist=8.75,
                    num_cpu=1):

    if out_path is None:
        out_path = f"{os.getcwd()}/{rascore_str}_{classify_str}"

    cluster_path = f"{get_neighbor_path(__file__,pipelines_str,data_str)}/{rascore_str}_{cluster_str}"

    file_path_lst = type_lst(file_paths)

    df = pd.DataFrame()
    for file_path in tqdm(
        file_path_lst,
        desc="Loading Files",
        position=0,
        leave=True,
    ):
        if type(file_path) == pd.DataFrame:
            temp_df = file_path.copy(deep=True)
            temp_df_col_lst = list(temp_df.columns)
            if modelid_col not in temp_df_col_lst:
                temp_df[modelid_col] = 0
        else:
            temp_df = pd.DataFrame()
            load_path_lst = list()

            if ".pdb" not in file_path and ".cif" not in file_path:
                temp_df = load_table(file_path)
                temp_df_col_lst = list(temp_df.columns)
                if core_path_col in temp_df_col_lst and chainid_col in temp_df_col_lst:
                    if modelid_col not in temp_df_col_lst:
                        temp_df[modelid_col] = 0
                else:
                    load_path_lst = load_lst(file_path)
                    temp_df = pd.DataFrame()
            else:
                load_path_lst = list([file_path])

            if len(temp_df) == 0 and len(load_path_lst) > 0:
                i = 0
                for load_path in tqdm(
                    load_path_lst,
                    desc="Loading Files",
                    position=1,
                    leave=True,
                ):
                    structure = load_coord(load_path)
                    for model in structure:
                        modelid = get_modelid(model)
                        for chain in model:
                            chainid = get_chainid(chain)
                            temp_df.at[i, core_path_col] = load_path
                            temp_df.at[i, modelid_col] = modelid
                            temp_df.at[i, chainid_col] = chainid
                            i += 1

        df = pd.concat([df, temp_df], sort=False)

    if len(df) == 0:
        print('No structures to classify.')
    else:
        df = df.reset_index(drop=True)

        df_col_lst = list(df.columns)

        for col in [y32_name, y71_name, sw1_name, sw2_name]:
            if col in df_col_lst:
                del df[col]

        missing_col_lst = [
            x for x in [core_path_col, modelid_col, chainid_col] if x not in df_col_lst
        ]
        if len(missing_col_lst) > 0:
            print(f"Input table missing columns: {lst_to_str(missing_col_lst)}")

        else:
            coord_path_lst = lst_col(df, core_path_col)

            if id_col not in df_col_lst and pdb_id_col not in df_col_lst:
                for index in list(df.index.values):
                    df.at[index, id_col] = get_file_name(df.at[index, core_path_col])

            if len([x for x in lig_col_lst if x not in df_col_lst]) > 0:
                df = annot_lig(
                    df=df,
                    site_dict=pharm_site_dict,
                    num_cpu=num_cpu,
                )

            if nuc_class_col not in df_col_lst:
                df[nuc_class_col] = df[bio_lig_col].map(nuc_class_dict).fillna(gtp_name)

            if dih_dict is None:
                dih_dict = prep_dih(coord_path_lst, num_cpu=num_cpu)

            dist_df = build_dist_table(
                df,
                x_resids=[y32_resid, y71_resid],
                y_resids=[g12_resid, v9_resid],
                x_atomids=['OH','OH'],
                y_atomids=['CA','CA'],
                atom_dist_col_lst=[y32_name, y71_name]
            )

            dist_dict = {y32_name: y32_dist, y71_name: y71_dist}

            for resid_name in [y32_name, y71_name]:
                dist_df[resid_name] = dist_df[resid_name].map(str)

            for index in list(dist_df.index.values):

                nuc_class = dist_df.at[index, nuc_class_col]

                for resid_name in [y32_name, y71_name]:

                    atom_dist = float(dist_df.at[index, resid_name])

                    if atom_dist == 999.00:
                        group_name = disorder_name
                    else:
                        group_name = resid_name
                        if atom_dist <= dist_dict[resid_name]:
                            group_name += in_name
                        else:
                            group_name += out_name

                    dist_df.at[index, resid_name] = group_name

            for resid_name in [y32_name, y71_name]:

                merge_df = dist_df.loc[
                    :, [core_path_col, modelid_col, chainid_col, resid_name]
                ]

                df = merge_tables(df, merge_df)

            for resid_name in [y32_name, y71_name]:

                if resid_name == y32_name:
                    loop_name = sw1_name
                    loop_resids = sw1_resids
                elif resid_name == y71_name:
                    loop_name = sw2_name
                    loop_resids = sw2_resids

                cluster_loop_path = f"{cluster_path}/{loop_name}"
                classify_loop_path = f"{out_path}/{loop_name}"

                loop_result_df = pd.DataFrame()
                loop_sum_df = pd.DataFrame()

                for group_name in lst_col(df, resid_name, unique=True):

                    group_df = mask_equal(df, resid_name, group_name)

                    for nuc_class in nuc_class_lst:

                        group_nuc_df = mask_equal(group_df, nuc_class_col, nuc_class)

                        group_nuc_name = f"{group_name}.{nuc_class}"

                        print(
                                f"Classifying {loop_name} conformations in {group_name}.{nuc_class} structures."
                            )

                        cluster_table_path = get_file_path(
                            cluster_table_file,
                            dir_str=group_nuc_name,
                            dir_path=cluster_loop_path,
                        )

                        fit_matrix_path = get_file_path(
                            dih_fit_matrix_file,
                            dir_str=group_nuc_name,
                            dir_path=cluster_loop_path,
                        )

                        result_table_path = get_file_path(
                            result_table_file,
                            dir_str=group_nuc_name,
                            dir_path=classify_loop_path,
                        )

                        pred_matrix_path = get_file_path(
                            pred_matrix_file,
                            dir_str=group_nuc_name,
                            dir_path=classify_loop_path,
                        )

                        if len(group_nuc_df) > 0:

                            dih_df = build_dih_table(
                                    df=group_nuc_df,
                                    dih_dict=dih_dict,
                                    bb_resids=loop_resids,
                                )

                            if disorder_name not in group_nuc_name:

                                cluster_df = load_table(cluster_table_path)

                                build_dih_matrix(
                                    fit_df=cluster_df,
                                    pred_df=dih_df,
                                    max_norm_path=pred_matrix_path,
                                )

                                fit_matrix = load_matrix(fit_matrix_path)
                                pred_matrix = load_matrix(pred_matrix_path)

                                classify_matrix(
                                    cluster_df=cluster_df,
                                    pred_df=dih_df,
                                    fit_matrix=fit_matrix,
                                    pred_matrix=pred_matrix,
                                    result_table_path=result_table_path,
                                    max_nn_dist=0.45,
                                    only_save_pred=True,
                                    reorder_class=False,
                                )
                            else:
                                result_df = dih_df.copy(deep=True)
                                result_df[cluster_col] = noise_name
                                save_table(result_table_path, result_df)

                            result_df = load_table(result_table_path)

                            loop_result_df = pd.concat(
                                [loop_result_df, result_df], sort=False
                            )
                       
                loop_result_table_path = get_file_path(
                    result_table_file, dir_str=loop_name, dir_path=out_path
                )
                loop_sum_table_path = get_file_path(
                    sum_table_file, dir_str=loop_name, dir_path=out_path
                )

                if pdb_id_col in list(loop_result_df.columns):

                    cluster_result_df = load_table(
                        get_file_path(
                            result_table_file,
                            dir_str=loop_name,
                            dir_path=cluster_path,
                        )
                    )

                    cluster_dict = make_dict(
                        lst_col(cluster_result_df, pdb_id_col),
                        lst_col(cluster_result_df, cluster_col),
                    )

                    result_dict = make_dict(
                        lst_col(loop_result_df, pdb_id_col),
                        lst_col(loop_result_df, cluster_col),
                    )

                    for pdb_id, cluster in cluster_dict.items():
                        if pdb_id in list(result_dict.keys()):
                            result_dict[pdb_id] = cluster

                    loop_result_df[cluster_col] = loop_result_df[pdb_id_col].map(
                        result_dict
                    )

                loop_result_df = loop_result_df.reset_index(drop=True)
            
                for index in list(loop_result_df.index.values):
                    cluster = loop_result_df.at[index, cluster_col]
                    if noise_name in cluster:
                        complete =  str(loop_result_df.at[index, complete_col])
                        if complete == str(True):
                            cluster = outlier_name
                        elif complete == str(False):
                            cluster = disorder_name
                        loop_result_df.at[index,cluster_col] = cluster

                loop_sum_df = build_sum_table(loop_result_df)

                save_table(loop_result_table_path, loop_result_df)
                save_table(loop_sum_table_path, loop_sum_df)

                loop_result_df = loop_result_df.rename(columns={cluster_col: loop_name})

                loop_result_df = loop_result_df.loc[:,[core_path_col, modelid_col, chainid_col, loop_name]]

                df = merge_tables(df, loop_result_df)

            if gtp_name in lst_col(df, nuc_class_col, unique=True):
                dist_df = build_dist_table(
                    mask_equal(df, nuc_class_col, gtp_name),
                    x_resids=[y32_resid],
                    y_resids=[bio_lig_col],
                    x_atomids=["OH"],
                    y_atomids=[gtp_atomids],
                    hb_status_col_lst=[hb_status_col],
                    check_hb=True,
                )

                merge_df = dist_df.loc[
                    :, [core_path_col, modelid_col, chainid_col, hb_status_col]
                ]

                df = merge_tables(df, merge_df)

            result_table_path = get_file_path(result_table_file, dir_path=out_path)

            for col in list(df.columns):
                df = fix_col(df, col)

            save_table(result_table_path, df)

            for loop_name in [sw1_name,sw2_name]:

                pymol_pml_path = get_file_path(
                    f"{loop_name}_{pymol_pml_file}", dir_path=out_path
                )

                if loop_name == sw1_name:
                    loop_resids = sw1_resids
                    stick_resids = [y32_resid]
                elif loop_name == sw2_name:
                    loop_resids = sw2_resids
                    stick_resids = [y71_resid]

                write_pymol_script(
                    df,
                    pymol_pml_path,
                    group_col=loop_name,
                    stick_resids=stick_resids,
                    loop_resids=loop_resids,
                    style_ribbon=True,
                    thick_bb=False,
                    show_bio=True,
                    sup_group=True,
                    color_palette=conf_color_dict[loop_name],
                    sup_resids=sup_resids,
                    show_resids=show_resids,
                )

    print("Rascore classification complete!")


# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import streamlit as st
from random import randint

from ..functions.col import id_col, core_path_col, pharm_lig_site_col, pocket_class_col
from ..functions.path import (
    delete_path,
    get_file_path,
    get_file_name,
    get_neighbor_path,
    load_table,
    get_dir_path,
    pages_str,
    data_str,
    rascore_str,
    classify_str,
)
from ..functions.lst import res_to_lst
from ..functions.file import result_table_file
from ..functions.gui import (
    save_st_file,
    show_st_dataframe,
    download_st_df,
    rename_st_cols,
    write_st_end,
)
from ..pipelines.classify_rascore import classify_rascore


def classify_page():

    st.markdown("## Classify Structures")

    st.markdown("---")

    st_file_lst = st.file_uploader(
        "Upload Coordinate Files", accept_multiple_files=True
    )

    table_st_file = None
    with st.expander("Optional Input", expanded=False):

        left_col, right_col = st.columns(2)

        sw1_resids = left_col.text_input('SW1 Residues',value="25-40")
        sw2_resids = right_col.text_input('SW2 Residues',value="56-76")

        sw1_len = len(res_to_lst('25-40'))
        sw2_len = len(res_to_lst('57-76'))

        if len(res_to_lst(sw1_resids)) != sw1_len:
            left_col.warning(f'SW1 Must Be {sw1_len} Residue Long')

        if len(res_to_lst(sw2_resids)) != len(res_to_lst('56-76')):
            right_col.warning(f'SW2 Must Be {sw2_len} Residue Long')

        y32_resid = int(left_col.text_input('Y32 Residue',value=32))
        y71_resid = int(right_col.text_input('Y71 Residue',value=71))

        g12_resid = int(left_col.text_input('G12 Residue',value=12))
        v9_resid = int(right_col.text_input('V9 Residue',value=9))

        st.markdown(
            """
        Tab-separated table with columns:
        - **core_path:** coordinate path:
        - **modelid:** model number (*optional*)
        - **chainid:** chain identifier
        - **nucleotide_class:** nucleotide state (*optional*)
        """
        )
        table_st_file = st.file_uploader(
            "Upload Table File", accept_multiple_files=False
        )

    with st.form(key="Classify Form"):
        out_file = st.text_input(
            label="Classify File Name", value="rascore_classify.tsv"
        )
        submit_files = st.form_submit_button(label="Classify Structures")

    if submit_files:
        if len(st_file_lst) > 0:
            with st.spinner(text="Classifying Structures"):

                try:
                    coord_path_lst = list()
                    id_dict = dict()
                    path_dict = dict()
                    for st_file in st_file_lst:
                        coord_path = save_st_file(st_file)
                        coord_path_lst.append(coord_path)
                        id_dict[get_file_name(coord_path)] = st_file.name
                        path_dict[st_file.name] = coord_path

                    classify_path = get_dir_path(
                        dir_str=f"{rascore_str}_{classify_str}_{randint(0,3261994)}",
                        dir_path=get_neighbor_path(__file__, pages_str, data_str),
                    )

                    if table_st_file is None:
                        classify_input = coord_path_lst
                    else:
                        table_path = save_st_file(table_st_file)
                        classify_input = load_table(table_path)
                        classify_input[core_path_col] = classify_input[
                            core_path_col
                        ].map(path_dict)
                        delete_path(table_path)

                    classify_rascore(classify_input, y32_resid=y32_resid, y71_resid=y71_resid, 
                    g12_resid=g12_resid,v9_resid=v9_resid,
                    sw1_resids=sw1_resids, sw2_resids=sw2_resids, 
                    out_path=classify_path)

                    st.success("Classified Structures")

                    result_file_path = get_file_path(
                        result_table_file, dir_path=classify_path
                    )

                    df = load_table(result_file_path)

                    df[id_col] = df[id_col].map(id_dict)

                    df = df.rename(columns={pharm_lig_site_col: pocket_class_col})

                    df = rename_st_cols(df)

                    show_st_dataframe(df)

                    download_st_df(df, out_file, "Download Classification Table")
                except:
                    st.error("Error Analyzing Uploaded Structures")

                delete_path(classify_path)
                for coord_path in coord_path_lst:
                    delete_path(coord_path)

    write_st_end()



from src.download.modules import *


def my_hook(t):
    last_b = [0]

    def update_to(b=1, b_size=1, t_size=None):
        if t_size is not None:
            t.total = t_size
        t.update((b - last_b[0]) * b_size)
        last_b[0] = b
    return update_to


def downloader_for_catalog_with_urllib(ftp_to_download, where_the_file_goes):
    socket.setdefaulttimeout(600)
    for _ in range(10):
        try:
            last_slash = ftp_to_download.rsplit('/', 1)[-1]
            with tqdm.tqdm(unit="B", unit_scale=True, desc="Downloading mmCIF/SIFTS catalogs " + last_slash, position=0, leave=True) as t:
                reporthook = my_hook(t)
                urllib.request.urlretrieve(ftp_to_download, Path(str(where_the_file_goes) + "/" + last_slash), reporthook=reporthook)
            break
        except Exception:
            time.sleep(1)


def catalog_downloader():
    """PDB ls-lR catalog"""
    ftp_for_all_mmCIF_and_PDB = "ftp://ftp.rcsb.org/pub/pdb/data/structures/ls-lR"

    ftp_to_download = ftp_for_all_mmCIF_and_PDB
    last_slash = ftp_to_download.rsplit('/', 1)[-1]
    today_date = date.today()
    today_date_str = today_date.strftime("_%Y_%m_%d")

    where_the_file_goes = current_directory + "/" + last_slash + today_date_str
    if not os.path.exists(where_the_file_goes):
        os.makedirs(where_the_file_goes)
        downloader_for_catalog_with_urllib(ftp_to_download, where_the_file_goes)
    else:
        if not os.path.isfile(where_the_file_goes + "/ls-lR"):
            downloader_for_catalog_with_urllib(ftp_to_download, where_the_file_goes)

    # downloader_for_catalog_with_urllib(ftp_to_download, where_the_file_goes)

    # reading txt file SIFTS parent catalog and creating pandas df out of it
    df_catalog_listing_everything = pd.read_csv(Path(str(where_the_file_goes) + "/" + last_slash),
                                                names=["1", "2", "3", "4", "Data_size", "Month", "Day", "Time",
                                                       "file_name", "10", "file_names_path"], sep="\s+",
                                                low_memory=False)

    # Dropping all unnecessary rows leaving only files with 'cif.gz' endings
    df_mmCIF_catalog_dropna = df_catalog_listing_everything.dropna()
    df_mmCIF_catalog_dropna_cif_gz = df_mmCIF_catalog_dropna[df_mmCIF_catalog_dropna['file_name'].str.endswith('cif.gz')]
    df_mmCIF_catalog_dropna_cif_gz_34kb = df_mmCIF_catalog_dropna_cif_gz[df_mmCIF_catalog_dropna_cif_gz.Data_size == 34.0]

    # Dropping all unnecessary rows leaving only files with 'ent.gz' endings
    df_PDB_catalog_dropna = df_catalog_listing_everything.dropna()
    df_PDB_catalog_dropna_ent_gz = df_PDB_catalog_dropna[df_PDB_catalog_dropna['file_name'].str.endswith('ent.gz')]
    df_PDB_catalog_dropna_ent_gz_35kb = df_PDB_catalog_dropna_ent_gz[df_PDB_catalog_dropna_ent_gz.Data_size == 35.0]

    # creating lists of the mmCIF file_names
    list_of_mmCIF_cif_gz_file_names = list()
    for mmCIF_file_name in df_mmCIF_catalog_dropna_cif_gz_34kb["file_name"]:
        list_of_mmCIF_cif_gz_file_names.append(mmCIF_file_name)

    # creating lists of the PDB file_names
    list_of_PDB_ent_gz_file_names = list()
    for PDB_file_name in df_PDB_catalog_dropna_ent_gz_35kb["file_name"]:
        list_of_PDB_ent_gz_file_names.append(PDB_file_name)

    """SIFTS xml catalog"""
    ftp_all_SIFTS = "ftp://ftp.ebi.ac.uk/pub/databases/msd/sifts/xml"
    ftp_to_download = ftp_all_SIFTS
    last_slash = ftp_to_download.rsplit('/', 1)[-1]

    where_the_file_goes = current_directory + "/" + last_slash + today_date_str
    if not os.path.exists(where_the_file_goes):
        os.makedirs(where_the_file_goes)
        downloader_for_catalog_with_urllib(ftp_to_download, where_the_file_goes)
    else:
        if not os.path.isfile(where_the_file_goes + "/xml"):
            downloader_for_catalog_with_urllib(ftp_to_download, where_the_file_goes)

    # downloader_for_catalog_with_urllib(ftp_to_download, where_the_file_goes)

    # reading txt file SIFTS parent catalog and creating pandas df out of it
    df_catalog_listing_everything = pd.read_csv(Path(str(where_the_file_goes) + "/" + last_slash),
                                                names=["1", "2", "3", "4", "Data_size", "Month", "Day", "Time",
                                                       "file_name", "10", "file_names_path"], sep="\s+",
                                                low_memory=False)

    # Dropping all unnecessary rows leaving only files with 'xml.gz' endings
    df_SIFTS_catalog_dropna = df_catalog_listing_everything.dropna()
    df_SIFTS_catalog_dropna_xml_gz = df_SIFTS_catalog_dropna[
        df_SIFTS_catalog_dropna['file_name'].str.endswith('xml.gz')]
    df_SIFTS_catalog_dropna_xml_gz_27kb = df_SIFTS_catalog_dropna_xml_gz[
        df_SIFTS_catalog_dropna_xml_gz.Data_size == 27.0]

    # creating lists of the SIFTS file_names
    list_of_SIFTS_xml_gz_file_names = list()
    for SIFTS_file_names in df_SIFTS_catalog_dropna_xml_gz_27kb["file_name"]:
        list_of_SIFTS_xml_gz_file_names.append(SIFTS_file_names)

    _4Char_list_of_SIFTS_xml_gz_file_names = list()
    for SIFTS_file_names_4Char in list_of_SIFTS_xml_gz_file_names:
        _4Char_list_of_SIFTS_xml_gz_file_names.append(SIFTS_file_names_4Char[:4])

    _4Char_list_of_PDB_ent_gz_file_names = list()
    for PDB_file_names_4Char in list_of_PDB_ent_gz_file_names:
        _4Char_list_of_PDB_ent_gz_file_names.append(PDB_file_names_4Char[3:7])

    _4Char_list_of_mmCIF_cif_gz_file_names = list()
    for mmCIF_file_names_4Char in list_of_mmCIF_cif_gz_file_names:
        _4Char_list_of_mmCIF_cif_gz_file_names.append(mmCIF_file_names_4Char[:4])

    df_list_of_mmCIF_cif_gz_file_names = pd.DataFrame(
        zip(list_of_mmCIF_cif_gz_file_names, _4Char_list_of_mmCIF_cif_gz_file_names), columns=["mmCIF", "4mmCIF"])
    df_list_of_PDB_ent_gz_file_names = pd.DataFrame(
        zip(list_of_PDB_ent_gz_file_names, _4Char_list_of_PDB_ent_gz_file_names), columns=["PDB", "4PDB"])
    df_list_of_SIFTS_xml_gz_file_names = pd.DataFrame(
        zip(list_of_SIFTS_xml_gz_file_names, _4Char_list_of_SIFTS_xml_gz_file_names), columns=["SIFTS", "4SIFTS"])

    merged_df_mmCIF_PDB_file_names = df_list_of_mmCIF_cif_gz_file_names.merge(df_list_of_PDB_ent_gz_file_names,
                                                                              left_on='4mmCIF', right_on='4PDB',
                                                                              how="left")
    merged_df_mmCIF_PDB_SIFTS_file_names = merged_df_mmCIF_PDB_file_names.merge(df_list_of_SIFTS_xml_gz_file_names,
                                                                                left_on='4mmCIF', right_on='4SIFTS',
                                                                                how="left")

    merged_df_mmCIF_PDB_SIFTS_file_names['SIFTS'] = merged_df_mmCIF_PDB_SIFTS_file_names['SIFTS'].replace(np.nan, "0000")
    merged_df_mmCIF_PDB_SIFTS_file_names['PDB'] = merged_df_mmCIF_PDB_SIFTS_file_names['PDB'].replace(np.nan, "0000")

    SIFTS_file_names_with_null_if_files_absent = list()
    for SIFTS_file_name_null_for_absent in merged_df_mmCIF_PDB_SIFTS_file_names['SIFTS']:
        SIFTS_file_names_with_null_if_files_absent.append(SIFTS_file_name_null_for_absent)

    PDB_file_names_with_null_if_files_absent = list()
    for PDB_file_name_null_for_absent in merged_df_mmCIF_PDB_SIFTS_file_names['PDB']:
        PDB_file_names_with_null_if_files_absent.append(PDB_file_name_null_for_absent)

    mmCIF_file_names_with_null_if_files_absent = list()
    for mmCIF_file_name_null_for_absent in merged_df_mmCIF_PDB_SIFTS_file_names['mmCIF']:
        mmCIF_file_names_with_null_if_files_absent.append(mmCIF_file_name_null_for_absent)

    return (mmCIF_file_names_with_null_if_files_absent,
            PDB_file_names_with_null_if_files_absent,
            SIFTS_file_names_with_null_if_files_absent)


"""usage of catalog_downloader()"""
# all_data_tuple = catalog_downloader()
# mmCIF_file_names_with_null_if_files_absent = all_data_tuple[0]
# PDB_file_names_with_null_if_files_absent = all_data_tuple[1]
# SIFTS_file_names_with_null_if_files_absent = all_data_tuple[2]

# print(len(mmCIF_file_names_with_null_if_files_absent))
# print(len(PDB_file_names_with_null_if_files_absent))
# print(len(SIFTS_file_names_with_null_if_files_absent))



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from Bio import SeqIO, pairwise2
from Bio.Seq import Seq
from Bio.Align import substitution_matrices
from Bio.SeqRecord import SeqRecord
from Bio.Align.Applications import ClustalOmegaCommandline

from .path import delete_path


def load_record_lst(path):

    return SeqIO.parse(path, "fasta")


def load_record_dict(path):

    return SeqIO.to_dict(load_record_lst(path))


def get_record_id(record):

    return str(record.id)


def get_record_seq(record):

    return str(record.seq)


def get_record_desc(record):

    return record.description


def build_record(seq, id, name=None, desc=None):

    return SeqRecord(Seq(seq), id=id, name=name, description=desc)


def clustal_omega(fasta_path, msa_path):

    delete_path(msa_path)

    cline = ClustalOmegaCommandline(infile=fasta_path, outfile=msa_path)
    cline()


def pair_seq_aln(
    ref_seq,
    mob_seq,
    **kwargs,
):

    matrix = kwargs.get("matrix", substitution_matrices.load("BLOSUM62"))
    gap_open = kwargs.get("gap_open", -10.0)
    gap_extend = kwargs.get("gap_extend", -0.5)

    return pairwise2.align.globalds(
        ref_seq, mob_seq, matrix, gap_open, gap_extend, penalize_end_gaps=(False, False)
    )


def calc_seq_id(ref_seq, mob_seq, gap=True, aln=True):

    if aln:
        aln = pair_seq_aln(ref_seq, mob_seq)

        best_aln = aln[0]

        ref_seq = best_aln[0]
        mob_seq = best_aln[1]

    seq_len = min(len(ref_seq), len(mob_seq))
    match_lst = [ref_seq[i] == mob_seq[i] for i in range(seq_len)]

    if gap:
        seq_id = (100 * sum(match_lst)) / seq_len
    else:
        gap_less_seq_len = sum(
            [1 for i in range(seq_len) if (ref_seq[i] != "-" and mob_seq[i] != "-")]
        )
        seq_id = (100 * sum(match_lst)) / gap_less_seq_len

    return seq_id



# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import os
import subprocess
import io

import pandas as pd

from tqdm import tqdm
import concurrent.futures
import pymol2

from ..functions.lst import (
    sort_lst,
    lst_unique,
    lst_to_str,
    type_lst,
)
from ..functions.table import (
    mask_equal,
    title_str,
    merge_dicts,
)
from ..functions.path import (
    delete_path,
    get_file_path,
    get_dir_path,
    get_file_name,
    copy_path,
    get_pocket_path,
    append_path,
    save_json,
    pocket_str,
)
from ..functions.coord import (
    load_coord,
    get_reschainid,
    get_residues,
    get_resnum,
    is_aa,
)
from ..functions.col import (
    pocket_path_col,
    pocket_cont_col,
    pocket_status_col,
    pocket_score_col,
    pocket_volume_col,
    pocket_type_col,
    pocket_lig_col,
    pharm_lig_col,
)
from ..functions.lig import lig_col_lst, lig_lst_dict

pocket_bound_name = "Bound"
pocket_unbound_name = "Unbound"

pocket_info_dict = {
    "Druggability Score": "drug_score",
    "Volume": "pock_vol",
    "Number of Alpha Spheres": "nb_AS",
}


def run_fpocket(
    coord_path,
    pocket_dir=None,
    chainid_lst=None,
):

    pocket_dir_path = get_dir_path(dir_str=pocket_str, dir_path=pocket_dir)

    coord_name = get_file_name(coord_path)
    coord_name = coord_name.replace("_core", "")

    run_path = get_file_path(coord_name, dir_str=pocket_str, dir_path=pocket_dir)

    copy_path(coord_path, run_path)

    cmd_lst = [
        "fpocket",
        "-f",
        run_path,
        "-d",
    ]

    df = pd.read_csv(
        io.StringIO(
            subprocess.Popen(cmd_lst, stdout=subprocess.PIPE)
            .communicate()[0]
            .decode("utf-8")
        ),
        delim_whitespace=True,
        dtype=str,
    )

    if chainid_lst is not None:
        for chainid in chainid_lst:
            df = mask_equal(df, "name_chain_1", chainid)
            df = mask_equal(df, "name_chain_2", chainid)

    pocket_dict = {coord_path: dict()}

    obj = coord_name.replace(".pdb", "")

    run_dir_path = f"{pocket_dir_path}/pocket_{obj}_out"

    with pymol2.PyMOL() as pymol:

        cmd = pymol.cmd

        cmd.load(coord_path, obj)

        for i, index in enumerate(list(df.index.values)):

            pocket = str(i + 1)

            pocket_id = df.at[index, "cav_id"]

            pocket_chainid_lst = lst_unique(
                (df.at[index, "name_chain_1"], df.at[index, "name_chain_2"])
            )

            pocket_chainid = lst_to_str(
                pocket_chainid_lst,
                join_txt="_",
            )

            cont_path = get_file_path(
                f"pocket{pocket_id}_atm.pdb", dir_path=run_dir_path
            )
            sphere_path = get_file_path(
                f"pocket{pocket_id}_vert.pqr", dir_path=run_dir_path
            )

            sphere_obj = f"sphere_{pocket}"
            cmd.load(sphere_path, sphere_obj)

            cmd.alter(sphere_obj, f"chain='{pocket_chainid}'")

            pocket_obj = f"pocket_{pocket}"
            cmd.create(pocket_obj, f"{obj} {sphere_obj}")

            pocket_path = get_pocket_path(
                obj,
                pocket,
                dir_path=pocket_dir,
            )
            cmd.save(pocket_path, pocket_obj)

            structure = load_coord(cont_path)

            pocket_residue_lst = get_residues(structure)

            if chainid_lst is not None:
                pocket_residue_lst = [
                    x for x in pocket_residue_lst if get_reschainid(x) in chainid_lst
                ]

            pocket_cont_lst = sort_lst(
                lst_unique(
                    [
                        get_resnum(x)
                        for x in pocket_residue_lst
                        if (get_resnum(x) < 50000) and is_aa(x)
                    ]
                )
            )

            pocket_lig = df.at[index, "lig_het_tag"]

            pocket_status = pocket_unbound_name
            pocket_type = pocket_unbound_name
            if pd.isna(pocket_lig):
                pocket_lig = "STP"
            else:
                pocket_status = pocket_bound_name
                is_pharm = True
                for lig_col in lig_col_lst:
                    if pocket_lig in lig_lst_dict[lig_col]:
                        pocket_type = lig_col
                        is_pharm = False
                if is_pharm:
                    pocket_type = pharm_lig_col
                pocket_type = title_str(pocket_type)

            pocket_dict[coord_path][pocket] = dict()

            pocket_dict[coord_path][pocket][pocket_volume_col] = df.at[index, "volume"]
            pocket_dict[coord_path][pocket][pocket_score_col] = df.at[
                index, "drug_score"
            ]
            pocket_dict[coord_path][pocket][pocket_status_col] = pocket_status
            pocket_dict[coord_path][pocket][pocket_type_col] = pocket_type
            pocket_dict[coord_path][pocket][pocket_lig_col] = pocket_lig
            pocket_dict[coord_path][pocket][pocket_cont_col] = lst_to_str(
                pocket_cont_lst
            )
            pocket_dict[coord_path][pocket][pocket_path_col] = pocket_path

    delete_path(run_path)
    delete_path(run_dir_path)

    return pocket_dict


def prep_pocket(
    coord_paths,
    pocket_dir=None,
    pocket_json_path=None,
    chainid_dict=None,
    update_pocket=False,
    num_cpu=1,
):

    pocket_dir_path = get_dir_path(dir_str=pocket_str, dir_path=pocket_dir)

    if update_pocket:
        delete_path(pocket_dir_path)
    append_path(pocket_dir_path)

    coord_path_lst = type_lst(coord_paths)

    pocket_dict = dict()

    if num_cpu == 1:
        for coord_path in tqdm(
            coord_path_lst,
            desc="Preparing pockets",
            position=0,
            leave=True,
        ):
            pocket_dict = merge_dicts(
                [
                    pocket_dict,
                    run_fpocket(
                        coord_path,
                        pocket_dir=pocket_dir,
                        chainid_lst=chainid_dict[coord_path],
                    ),
                ]
            )
    else:
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cpu) as executor:
            job_lst = [
                executor.submit(
                    run_fpocket,
                    coord_path,
                    pocket_dir=pocket_dir,
                    chainid_lst=chainid_dict[coord_path],
                )
                for coord_path in coord_path_lst
            ]

            for job in tqdm(
                concurrent.futures.as_completed(job_lst),
                desc="Preparing pockets",
                total=len(job_lst),
                miniters=1,
                position=0,
                leave=True,
            ):

                pocket_dict = merge_dicts([pocket_dict, job.result()])

    print("Prepared pockets!")

    if pocket_json_path is not None:
        save_json(pocket_json_path, pocket_dict)
    else:
        return pocket_dict


# -*- coding: utf-8 -*-
"""
  Copyright 2022 Mitchell Isaac Parker

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import pandas as pd
import numpy as np

import statsmodels.api as sm
from statsmodels.sandbox.stats.multicomp import multipletests
from scipy import stats

from .lst import type_lst
from .table import lst_col, mask_equal, mask_unequal
from .col import (
    p_col,
    correct_p_col,
    total_col,
    index_col,
    a_col,
    b_col,
    c_col,
    d_col,
    risk_ratio_col,
    up_ci_col,
    low_ci_col,
    sig_col,
    corr_col,
)


def label_sig(val, label="*", cutoff=0.05):

    if val > cutoff:
        return "ns"
    else:
        return label


def correct_p_vals(df, correct_method="fdr_bh"):

    df = df.reset_index(drop=True)

    p_lst = lst_col(df, p_col)

    correct_p_lst = multipletests(p_lst, method=correct_method)[1]

    for index in list(df.index.values):

        correct_p = correct_p_lst[index]

        df.at[index, correct_p_col] = correct_p

    return df


def calc_rr(df, exp_cols, out_cols, correct_method="fdr_bh"):

    df = df.dropna()
    df = df.reset_index(drop=True)

    i_col_lst = type_lst(out_cols)
    j_col_lst = type_lst(exp_cols)

    col_lst = i_col_lst + j_col_lst

    if total_col not in list(df.columns):

        df.index.name = index_col
        df = df.reset_index()

        df = df.groupby(col_lst)[index_col].nunique().reset_index(name=total_col)

    total = df[total_col].sum()

    df_index_lst = list(df.index.values)

    for index in df_index_lst:

        i_df = df.copy(deep=True)
        j_df = df.copy(deep=True)

        for col in i_col_lst:

            val = df.at[index, col]

            i_df = mask_equal(i_df, col, val)

        for col in j_col_lst:

            val = df.at[index, col]

            j_df = mask_equal(j_df, col, val)

        a = df.at[index, total_col]
        b = i_df[total_col].sum() - a
        c = j_df[total_col].sum() - a
        d = total - (a + b + c)

        matrix = np.zeros((2, 2))

        matrix[0, 0] = a
        matrix[0, 1] = b
        matrix[1, 0] = c
        matrix[1, 1] = d

        table = sm.stats.Table2x2(matrix)

        risk_ratio = table.riskratio
        p_val = table.riskratio_pvalue()
        low_ci = table.riskratio_confint()[0]
        up_ci = table.riskratio_confint()[1]

        df.at[index, a_col] = a
        df.at[index, b_col] = b
        df.at[index, c_col] = c
        df.at[index, d_col] = d
        df.at[index, risk_ratio_col] = risk_ratio
        df.at[index, p_col] = p_val
        df.at[index, low_ci_col] = low_ci
        df.at[index, up_ci_col] = up_ci

    if correct_method is not None:
        df = correct_p_vals(df, correct_method=correct_method)

    df = df.sort_values(by=col_lst)

    noise_df = pd.DataFrame()
    for col in col_lst:
        temcorr_df = mask_equal(df, col, "Noise")
        noise_df = pd.concat([noise_df, temcorr_df], sort=False)
        df = mask_unequal(df, col, "Noise")

    df = pd.concat([df, noise_df], sort=False)

    df = df.reset_index(drop=True)

    df[sig_col] = df[correct_p_col].map(label_sig)

    return df


def calc_corr_stat(df, x_col, y_col, return_df=False, use_kt=False):

    if len(df) < 2:
        corr = np.nan
        p_val = np.nan
    else:
        df[x_col] = df[x_col].map(float)
        df[y_col] = df[y_col].map(float)

        x_lst = lst_col(df, x_col)
        y_lst = lst_col(df, y_col)

        if use_kt:
            corr, p_val = stats.kendalltau(x_lst, y_lst)
        else:
            corr, p_val = stats.pearsonr(x_lst, y_lst)

    if return_df:
        return pd.DataFrame({corr_col: [corr], p_col: [p_val]})
    else:
        return (corr, p_val)


def calc_corr(df, x_col, y_col, hue_cols=None, correct_method="fdr_bh", use_kt=False):

    val_df = df.copy(deep=True)

    if hue_cols is None:
        corr_df = calc_corr_stat(val_df, x_col, y_col, return_df=True, use_kt=use_kt)
    else:
        hue_col_lst = type_lst(hue_cols)

        corr_df = val_df.copy(deep=True)

        corr_df.index.name = index_col
        corr_df = corr_df.reset_index()

        corr_df = (
            corr_df.groupby(hue_col_lst)[index_col]
            .nunique()
            .reset_index(name=total_col)
        )

        df_index_lst = list(corr_df.index.values)

        for index in df_index_lst:

            hue_df = val_df.copy(deep=True)

            for col in hue_col_lst:

                val = corr_df.at[index, col]

                hue_df = mask_equal(hue_df, col, val)

            corr = calc_corr_stat(hue_df, x_col, y_col, use_kt=use_kt)

            corr_df.at[index, corr_col] = corr[0]
            corr_df.at[index, p_col] = corr[1]

        if correct_method is not None:
            corr_df = correct_p_vals(corr_df, correct_method=correct_method)

    corr_df = corr_df.sort_values(by=hue_col_lst)

    corr_df = corr_df.reset_index(drop=True)

    corr_df[sig_col] = corr_df[correct_p_col].map(label_sig)

    return corr_df
